import pytest
import sidekick as sk
from numpy.testing import assert_almost_equal
from sidekick import X, Y

from pydemic.formulas import sir, seir, seair, formula


class TestFormulaDecorator:
    def test_positional(self):
        @formula((), positional=1)
        def fn(f, gamma, sigma):
            return f(gamma, sigma)

        assert fn((X + Y), gamma=1, sigma=2) == 3
        assert fn((X + Y), {"gamma": 1, "sigma": 2}) == 3
        assert fn((X + Y), {"gamma": 1, "sigma": 2, "R0": 3}) == 3
        assert fn((X + Y), {"gamma": 1, "incubation_period": 0.5, "beta": 3}) == 3

        with pytest.raises(TypeError):
            fn((X + Y), gamma=1)

        with pytest.raises(TypeError):
            fn((X + Y), gamma=1, sigma=2, beta=3)

    def test_options(self):
        @formula((), positional=1, options=("power",))
        def fn(f, gamma, sigma, power=1):
            return f(gamma, sigma) ** power

        assert fn((X + Y), gamma=1, sigma=2) == 3
        assert fn((X + Y), gamma=1, sigma=2, power=2) == 9

    def test_ignore_kwargs(self):
        @formula((), invalid="pass")
        def fn(gamma, **kwargs):
            return gamma + sum(kwargs.values())

        assert fn(gamma=1, sigma=2) == 3
        assert fn(gamma=1, sigma=2, beta=3) == 6


class TestFormulaSIR:
    model = sir
    K = sk.delegate_to("model")
    R0 = sk.delegate_to("model")
    R0_from_K = sk.delegate_to("model")

    def test_R0(self):
        assert self.R0(beta=2, gamma=1) == 2
        assert self.R0.formula(beta=2, gamma=1) == 2
        assert self.R0.formula(2, 1) == 2
        assert self.R0(beta=2, infectious_period=2) == 4
        assert self.R0({"beta": 2, "gamma": 1}) == 2
        assert self.R0({"beta": 2, "gamma": 1}, gamma=2) == 1

    def test_R0_from_K(self):
        assert self.R0_from_K(K=0, gamma=1) == 1.0
        assert self.R0_from_K(K=0, gamma=2) == 1.0
        assert self.R0_from_K(K=1, gamma=2) == 1.5
        assert self.R0_from_K(K=2, gamma=2) == 2.0

    def test_K(self):
        assert self.K(R0=2, gamma=2) == 2.0
        assert self.K(R0=2, gamma=1) == 1.0
        assert self.K(R0=1, gamma=2) == 0.0
        assert self.K(R0=1, gamma=3) == 0.0
        assert self.K(R0=0.5, gamma=2) == -1.0
        assert self.K(R0=0.5, gamma=1) == -0.5

    def test_initial_state(self):
        assert_almost_equal(
            self.model.state_from_cases(population=10_000, cases=1000, R0=2, gamma=0.25),
            [9000, 333 + 2 / 3, 666 + 1 / 3],
        )


class TestFormulaSEIR(TestFormulaSIR):
    model = seir

    def test_K(self):
        assert self.K(R0=4, gamma=1, sigma=1) == 1.0
        assert self.K(R0=1, gamma=2, sigma=1) == 0.0
        assert self.K(R0=1, gamma=3, sigma=1) == 0.0
        assert self.K(R0=1 / 4, gamma=1, sigma=1) == -0.5

    def test_R0_from_K(self):
        assert self.R0_from_K(K=0, gamma=1, sigma=1) == 1.0
        assert self.R0_from_K(K=0, gamma=2, sigma=1) == 1.0
        assert self.R0_from_K(K=1, gamma=1, sigma=1) == 4.0
        assert self.R0_from_K(K=2, gamma=1, sigma=1) == 9.0

    def test_initial_state(self):
        assert_almost_equal(
            self.model.state_from_cases(
                population=10_000, cases=1000, R0=2, gamma=0.25, sigma=0.25
            ),
            [8705.9854608, 294.0145392, 207.8996744, 792.1003256],
        )


class TestFormulaSEAIR(TestFormulaSEIR):
    model = seair

    def test_R0(self):
        assert self.R0(beta=2, gamma=1, prob_symptoms=1.0, rho=0.25) == 2
        assert self.R0(beta=2, gamma=1, prob_symptoms=1.0, rho=0.5) == 2

        assert self.R0(beta=2, gamma=1, prob_symptoms=0.0, rho=0.25) == 0.5

        assert self.R0(beta=2, gamma=1, prob_symptoms=0.5, rho=1.0) == 2
        assert self.R0(beta=2, gamma=1, prob_symptoms=0.25, rho=1.0) == 2

        assert self.R0(beta=2, gamma=1, prob_symptoms=0.25, rho=0.0) == 0.5
        assert self.R0(beta=2, gamma=1, prob_symptoms=0.5, rho=0.0) == 1.0

        assert self.R0(beta=2, gamma=1, prob_symptoms=0.5, rho=0.5) == 1.5

    def test_initial_state(self):
        assert_almost_equal(
            self.model.state_from_cases(
                population=10_000,
                cases=1000,
                R0=2,
                gamma=0.25,
                sigma=0.25,
                prob_symptoms=0.5,
                rho=0.5,
            ),
            [9204.0712473, 588.0290783, 207.8996744, 207.8996744, 1584.2006512],
        )



import pandas as pd

from . import K
from .epidemic_curves import epidemic_curve
from .utils import cases
from .. import formulas
from ..diseases import disease as get_disease
from ..docs import docstring

ARGS = """Args:
    model ({'SIR', 'SEIR', 'SEAIR', etc}):
            Epidemic model used to compute R(t) from K(t)
    curves (pd.DataFrame):
        A dataframe with epidemic curves.
    window (int):
        Window size of triangular smoothing. Can be a single number or
        a 2-tuple with the window used to compute daily cases and the window
        used to smooth out the log-derivative.
    Re (bool):
        If True, return the effective reproduction number, instead of R0. Re
        does not consider the depletion of susceptibles and thus does not
        require to known the population size.
    population:
        Required parameter to obtain R(t), if model is passed as a string.
        This may be required if params is initialized from a disease object.
    params:
        Optional object holding simulation params. This argument is not necessary
        if model can be treated as Params or if a disease is given.
"""
METHODS_RT = {}
METHODS_R0 = {}
METHODS = {"Rt": METHODS_RT, "R0": METHODS_R0}


def method(which, names):
    """
    Decorator that register implementations to be used by the estimate_Rt()
    and estimate_R() functions.
    """
    names = [names] if isinstance(names, str) else names
    db = METHODS[which]

    def decorator(fn):
        for name in names:
            db[name] = fn
        return fn

    return decorator


@docstring(args=ARGS)
def estimate_Rt(model, curves: pd.DataFrame, method="RollingOLS", **kwargs) -> pd.DataFrame:
    """
    Estimate R(t) from epidemic curves and model.

    {args}

    Returns:
        A DataFrame with "Rt" and possibly other columns, depending on the
        method.

    See Also:
        naive_Rt
        rolling_OLS_Rt
    """
    return METHODS_RT[method](model, curves, **kwargs)


@docstring(args=ARGS)
@method("Rt", "naive")
def naive_Rt(model, curves: pd.DataFrame, window=14, **kwargs) -> pd.DataFrame:
    """
    Naive inference of R(t) from Epidemic curves using the naive_Kt() function.

    {args}


    See Also:
        :func:`pydemic.fitting.K.naive_Kt`
    """

    Kt = K.naive_Kt(curves, window=window)
    return Rt_from_Kt(model, curves, Kt, **kwargs)


@docstring(args=ARGS)
@method("Rt", ["RollingOLS", "ROLS"])
def rolling_OLS_Rt(model: str, curves: pd.DataFrame, window=14, **kwargs):
    """
    Compute R(t) from K(t) using the K.rolling_ols() function.

    {args}

    See Also:
        :func:`pydemic.fitting.K.rolling_OLS_Kt`
    """

    Kt = K.rolling_OLS_Kt(curves, window=window)
    return Rt_from_Kt(model, curves, Kt, **kwargs)


#
# Auxiliary functions
#
def Rt_from_Kt(model, curves, Kt, Re=False, **kwargs) -> pd.DataFrame:
    """
    Return Rt from Kt for model.

    Wraps a common logic for many functions in this module.
    """

    if isinstance(model, str):
        params = None
    else:
        params = model
        model = model.epidemic_model_name()

    params = kwargs.pop("params", params)
    if params is None:
        disease = get_disease()
        params = disease.params()

    re = pd.DataFrame(
        {col: formulas.R0_from_K(model, params, K=data, **kwargs) for col, data in Kt.iteritems()}
    )
    re.columns = ["Rt" + c[2:] if c.startswith("Kt") else c for c in re.columns]
    if Re:
        return re

    data = epidemic_curve(model, cases(curves), params)
    depletion = data["susceptible"] / data.sum(1)
    re /= depletion.values[:, None]
    return re



from typing import Union, Sequence, TypeVar

import numpy as np
import pandas as pd

Data = Union[Sequence, pd.Series, np.ndarray]
T = TypeVar("T", pd.Series, pd.DataFrame, np.ndarray, Sequence)


def trim_zeros(data: T, direction="both") -> T:
    """
    Remove null values from the start and end of series or array.
    """
    # TODO: use numpy's trim_zeros internally.
    if direction == "none":
        return data

    trim_left = direction in ("left", "both", "begin", "start")
    trim_right = direction in ("right", "both", "end", "finish")

    i = j = 0
    if np.ndim(data) > 1:
        values = data.values
        if trim_left:
            for i, x in enumerate(values):
                if not (x == 0).all():
                    break
        if trim_right:
            for j, x in enumerate(reversed(values)):
                if not (x == 0).all():
                    break
    else:
        if trim_left:
            for i, x in enumerate(data):
                if x != 0:
                    break
        if trim_right:
            for j, x in enumerate(reversed(data)):
                if x != 0:
                    break
    j = max(len(data) - j, i)

    if hasattr(data, "iloc"):
        return data.iloc[i:j]
    return data[i:j]


def force_monotonic(data: T, decreasing=False, check=True, copy=True) -> T:
    """
    Force series to be monotonically increasing/decreasing

    Args:
        data (pd.Series):
            Data to become monotonic.
        decreasing (bool):
            If True, force a monotonic decreasing behavior.
        copy:
            If False, make changes inplace.
        check:
            If False, prevent checking if data is monotonic before fixing
            its contents.

    Returns:
        pd.Series
    """

    if copy:
        data = data.copy()

    if isinstance(data, pd.DataFrame):
        for col in data:
            data[col] = force_monotonic(data[col], decreasing, check, copy=False)
        return data

    if check:
        if decreasing and data.is_monotonic_decreasing:
            return data
        elif not decreasing and data.is_monotonic_increasing:
            return data

    indexes = np.arange(len(data))
    idx = indexes[(data.diff() < 0).values]
    for i in reversed(idx):
        data.iloc[i - 1] = data.iloc[i]
    return data



from ..utils import format_args


class WrappedParams:
    """
    A wrapper object that exposes a disease as a param namespace.
    """

    __slots__ = ("_wrapped", "_args", "_kwargs", "_cache", "_keys", "_blacklist")

    def __init__(self, wrapped, *args, params_blacklist=None, **kwargs):
        self._wrapped = wrapped
        self._args = args
        self._kwargs = kwargs
        self._cache = {}

        if params_blacklist is None:
            params_blacklist = getattr(wrapped, "PARAMS_BLACKLIST", None)
        if params_blacklist is None:
            params_blacklist = getattr(wrapped, "params_blacklist", ())
        self._blacklist = frozenset(params_blacklist)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError(f"keys must be strings, got {item!r}")
        try:
            return self._fetch_param(item)
        except ValueError:
            raise KeyError(item)

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(item)
        try:
            return self._fetch_param(item)
        except ValueError:
            raise AttributeError(item)

    def __getstate__(self):
        return self._wrapped, self._args, self._kwargs

    def __setstate__(self, state):
        self._wrapped, self._args, self._args = state
        self._cache = {}

    def __repr__(self):
        args = format_args(self._wrapped, *self._args, **self._kwargs)
        return f"WrappedParams({args})"

    def __iter__(self):
        cls = type(self)
        blacklist = self._blacklist
        for attr in dir(self._wrapped):
            if attr.startswith("_") or hasattr(cls, attr) or attr in blacklist:
                continue

            value = getattr(self._wrapped, attr)
            if callable(value):
                yield attr

    def _fetch_param(self, name):
        """
        Obtain parameter with the given name and save result in cache.
        """
        try:
            return self._cache[name]
        except KeyError:
            try:
                value = getattr(self._wrapped, name)
            except AttributeError:
                raise ValueError(f"invalid parameter: {name!r}")
            if callable(value):
                value = value(*self._args, **self._kwargs)
            self._cache[name] = value
            return value



from abc import ABC
from typing import Callable, TYPE_CHECKING

import numpy as np
import pandas as pd
import sidekick as sk

from .data_transforms import DATA_TRANSFORMS, MODEL_TRANSFORMS

if TYPE_CHECKING:
    from .meta_info import Meta

ListLike = (list, np.ndarray, pd.Index)


class WithDataMixin(ABC):
    """
    Subclasses have a ".data" dataframe attribute that hold time-series
    information about a simulation.
    """

    meta: "Meta"
    data: pd.DataFrame
    times: pd.Index
    get_param: Callable[[str], float]

    def __init__(self, data=None):
        if data is not None:
            self.data = data

    def __getitem__(self, item):
        if isinstance(item, tuple):
            first, idx = item

            if isinstance(first, str):
                col, _, transform = first.rpartition(":")
                if col:
                    fn = self.data_transformer(transform)
                    return fn(col, idx)
                else:
                    col = transform
                    try:
                        return self.get_column(col, idx)
                    except ValueError:
                        raise KeyError(item)

            elif isinstance(first, ListLike):
                data = [self[col, idx] for col in first]
                return pd.concat(data, axis=1)

        elif isinstance(item, (str, *ListLike)):
            return self.__getitem__((item, None))

        elif isinstance(item, slice):
            raise NotImplementedError
        else:
            raise TypeError(f"invalid item: {item!r}")

    def get_column(self, name, idx):
        """
        Return a data column with the given name.
        """

        name = name.replace("-", "_")

        # Specialized data getter methods take precedence
        try:
            method = getattr(self, f"get_data_{name}")
        except AttributeError:
            pass
        else:
            return method(idx)

        # The next step is to check if requested column is in the state space
        name = self.meta.data_aliases.get(name, name)
        try:
            if idx is None:
                data = self.data[name]
            else:
                data = self.data[name].iloc[idx]
        except KeyError:
            pass
        else:
            return data.copy()

        # Finally, it may correspond to a parameter. We have two options. It may
        # be explicitly stored as a time-series or it may be computed implicitly
        # from the other parameters.
        if name in self.meta.params.all:
            x = self.get_param(name)
            times = self.times[idx or slice(None, None)]
            return pd.Series([x] * len(times), index=times, name=name)
        raise ValueError(f"invalid column: {name!r}")

    def data_transformer(self, name):
        """
        Return a transformer function that reads a column represented by name
        (usually calling self.get_data(col) and then transform the result by
        some operation.

        Default operations:
            dates, days, weeks, months, years:
                Convert index of the series or dataframe to days/weeks/months or
                years instead of dates.
            df, np:
                Force output to be a dataframe/numpy array, even when it has a
                single column.
            pp, ppc, p1k, p10k, p100k, p1m:
                Divide data and display it per population. The different names
                represent simple conversions: pp = data / population,
                pc = 100 * pp, p1k = 1000 pp, and so on.
            int, float, str:
                Coercion to integers, floats, or strings.
            round, round1, round2, round3:
                Rounding to 0, 1, 2, or 3 decimal places.
            human, pcfmt, p1kfmt:
                Human-friendly numeric formats.
        """

        if name in MODEL_TRANSFORMS:
            fn = MODEL_TRANSFORMS[name]
            return lambda col, idx: fn(self, col, idx)

        try:
            fn = DATA_TRANSFORMS[name]
            return lambda col, idx: fn(self[col], idx)

        except KeyError:
            raise ValueError(f"Invalid transform: {name}")


class WithDataModelMixin(WithDataMixin, ABC):
    """
    Mixin for simulation model subclasses with a data attribute.
    """

    def initialize(self):
        raise NotImplementedError

    @sk.lazy
    def data(self):
        self.initialize()
        return self.__dict__["data"]

    def get_data_population(self, idx):
        """
        Return current population.
        """
        return self.data.iloc[idx or slice(None, None)].sum(1)

    def get_data_N(self, idx):
        """
        Shorthand for get_data_population.
        """
        return self.get_data_population(idx)



from types import SimpleNamespace

from .param import Params


class ClinicalParams(Params):
    pass


#
# Clinical parameters
#
CLINICAL_DEFAULT = Params(
    "Default",
    hospitalization_period=7.0,
    icu_period=7.5,
    severe_delay=5.0,
    hospitalization_overflow_bias=0.25,
    critical_delay=7.0,
    prob_severe=0.18,
    prob_critical=0.05,
    prob_fatality=0.015 / 0.05,
    prob_no_hospitalization_fatality=0.30,
    prob_no_icu_fatality=1.00,
    case_fatality_ratio=0.015,
    infection_fatality_ratio=0.015 * 0.14,
    hospital_fatality_ratio=0.05,
)
clinical = SimpleNamespace(DEFAULT=CLINICAL_DEFAULT)



from collections import namedtuple
from typing import NamedTuple, Tuple, Iterable

import numpy as np
import pandas as pd

Result = namedtuple("Result", ["value", "info"])
ValueCI = namedtuple("ValueCI", ["value", "low", "high"])


class ValueStd(NamedTuple):
    """
    A value that represents a number with its standard deviation.
    """

    value: float
    std: float

    @classmethod
    def mean(cls, iterable: Iterable[Tuple[float, float]], tol=1e-9) -> "ValueStd":
        """
        Merge several independent point estimates of (value, std) into a single
        estimate, weighting results by the inverse variance.

        Args:
            iterable:
                A sequence of (value, std) tuples.
            tol:
                A normalization term to avoid problem with null variances.
        """
        if isinstance(iterable, pd.DataFrame):
            iterable = iterable.values

        weights = 0.0
        cum_var = 0.0
        N = 0

        for (value, std) in iterable:
            var = std * std + tol
            weight = 1 / var
            weights += weight
            cum_var += weight * value
            N += 1

        return ValueStd(cum_var / weights, np.sqrt(cum_var / N))


class ImproperlyConfigured(Exception):
    """
    Exception raised when trying to initialize object with improper configuration.
    """



from typing import Sequence

import numpy as np
import pandas as pd

from . import K
from .Rt import ARGS, METHODS_R0, method, rolling_OLS_Rt
from .epidemic_curves import epidemic_curve
from .utils import cases
from .. import formulas
from ..diseases import disease as get_disease
from ..docs import docstring
from ..types import ValueStd


@docstring(args=ARGS)
def estimate_R0(model, curves: pd.DataFrame, method="OLS", **kwargs) -> ValueStd:
    """
    Estimate R0 from epidemic curves and model.

    {args}

    Returns:
        A ValueStd with R0 and its associated standard deviation.

    See Also:
        naive_R0
        OLS_R0
    """
    return METHODS_R0[method](model, curves, **kwargs)


@docstring(args=ARGS)
@method("R0", "naive")
def naive_R0(model, curves: pd.DataFrame, window=14, **kwargs) -> ValueStd:
    """
    Naive inference of R(t) from Epidemic curves using the naive_Kt() function.

    {args}


    See Also:
        :func:`pydemic.fitting.K.naive_K`
    """

    k = K.naive_K(curves, window=window)
    return R0_from_K(model, curves, k, **kwargs)


@docstring(args=ARGS)
@method("R0", "OLS")
def ordinary_least_squares_R0(model: str, curves: pd.DataFrame, window=14, **kwargs) -> ValueStd:
    """
    Compute R0 from K using the K.ordinary_least_squares_K() function.

    {args}

    See Also:
        :func:`pydemic.fitting.K.ordinary_least_squares_K`
    """

    k = K.ordinary_least_squares_K(curves, window=window)
    return R0_from_K(model, curves, k, **kwargs)


@docstring(args=ARGS)
@method("R0", "RollingOLS")
def rolling_OLS_R0(model: str, curves: pd.DataFrame, window=14, **kwargs) -> ValueStd:
    """
    Compute R0 from K using the rolling_OLS_Rt() function and take the
    average value, weighting the most recent days with an exponential decay.

    {args}

    See Also:
        :func:`pydemic.fitting.K.ordinary_least_squares_K`
    """
    a, b = window if isinstance(window, Sequence) else (window, window)
    Rt = rolling_OLS_Rt(model, curves, window=window, **kwargs)
    weights = np.exp(np.arange(-len(Rt), 0) / b)
    R0, low, high = np.average(Rt.values, weights=weights, axis=0)
    return ValueStd(R0, (high - low) / 2)


#
# Auxiliary functions
#
def R0_from_K(model, curves, K, Re=False, **kwargs) -> ValueStd:
    """
    Return (R0, std) from K for model.

    Wraps a common logic for many functions in this module.
    """

    if isinstance(model, str):
        params = None
    else:
        params = model
        model = model.epidemic_model_name()

    params = kwargs.pop("params", params)
    if params is None:
        disease = get_disease()
        params = disease.params()

    mean, std = K
    K_mean = formulas.R0_from_K(model, params, K=mean, **kwargs)
    K_high = formulas.R0_from_K(model, params, K=mean + std, **kwargs)
    K_low = formulas.R0_from_K(model, params, K=mean - std, **kwargs)
    K_std = np.sqrt(((K_mean - K_high) ** 2 + (K_mean - K_low) ** 2) / 2)

    if Re:
        return ValueStd(K_mean, K_std)

    data = epidemic_curve(model, cases(curves), params)
    depletion = data["susceptible"] / data.sum(1)
    factor = 1 / depletion.iloc[-1]
    return ValueStd(K_mean * factor, K_std * factor)



from .abstract_seair import AbstractSEAIR


class eSEAIR(AbstractSEAIR):
    """
    A simple SEAIR model linearized around the DFE.
    """

    def run_to_fill(self, data, ts):
        raise NotImplementedError



import weakref
from collections import MutableMapping
from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Set

import pandas as pd

if TYPE_CHECKING:
    from ..models import Model


class BaseAttr(MutableMapping):
    """
    Base class for Info and Results objects.
    """

    model: "Model"
    _method_namespace: str = None
    __slots__ = ("_model_ref",)

    @property
    def _cache(self):
        raise NotImplementedError

    @property
    def model(self):
        return self._model_ref()

    def __init__(self, obj):
        self._model_ref = weakref.ref(obj)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[str]:
        seen = set()
        prefix = f"get_{self._method_namespace}_keys_"
        n = len(prefix)

        for method in dir(self.model):
            if method.startswith(prefix) and "__" not in method:
                key = method[n:]
                if key not in seen:
                    yield key
                seen.add(key)

        for key in self._cache:
            if key not in seen:
                yield key

    def __getitem__(self, item):
        group, key = normalize_key(item)
        cache = self._cache

        namespace = self._method_namespace
        model = self.model

        if key:
            cache = cache[group]
            try:
                return cache[key]
            except KeyError:
                pass
            result = get_scalar_item(model, group, key, namespace)

            # Save in cache only explicit keys
            # This avoid problems with aliases
            if key in get_dict_keys(model, group, namespace):
                cache[key] = result
            return result

        else:
            out = cache[group]
            try:
                out.update(get_dict_item(model, group, namespace, exclude=set(out)))
            except KeyError:
                if out:
                    return out
                raise
            return out

    def __setitem__(self, item, value):
        group, key = normalize_key(item)
        if key:
            self._set_item(group, key, value)
        else:
            for k, v in value.items():
                self._set_item(group, k, v)

    def __delitem__(self, item):
        group, key = normalize_key(item)
        if key:
            del self._cache[group][key]
        else:
            del self._cache[group]

    def _set_item(self, group, key, value):
        # Can be overridden by subclasses
        self._cache[group][key] = value

    def to_dict(self, flat=False):
        """
        Convert data to nested dict.
        """
        data = {}
        for k, v in self.items():
            if flat:
                data.update({f"{k}.{ki}": vi for ki, vi in v.items()})
            else:
                data[k] = v
        return data

    def to_frame(self) -> pd.DataFrame:
        """
        Expose information about object as a dataframe.
        """
        data = self.to_series()
        return pd.DataFrame({self._method_namespace: data})

    def to_series(self) -> pd.Series:
        """
        Expose information about object as a dataframe.
        """
        return pd.Series(self.to_dict(flat=True))

    def _html_repr_(self):
        return self.to_frame()._html_repr_()


def normalize_key(key):
    """
    Return tuple of (group, key) from key.
    """

    if isinstance(key, str):
        group, _, key = key.partition(".")
    elif isinstance(key, tuple):
        group, key = key
    else:
        raise TypeError(f"invalid key type: {type(key).__class__}")
    return group, key or None


def extra_keys(model, name, namespace):
    """
    Collect extra keys implemented as get_{namespace}_{group}__{name} methods.
    """

    prefix = f"get_{namespace}_value_{name}__"
    n = len(prefix)

    for attr in dir(model):
        if attr.startswith(prefix):
            yield attr[n:].replace("__", ".")


def get_scalar_item(model: "Model", group, key, namespace):
    """
    Fetch scalar value from Model instance attribute.
    """
    assert isinstance(namespace, str)

    name = f"get_{namespace}_value_{group}"
    full_name = f"{name}__{key}"

    if hasattr(model, full_name):
        fn = getattr(model, full_name)
        return fn()

    try:
        fn = getattr(model, name)
    except AttributeError:
        cls = type(model).__name__
        raise KeyError(f"{cls} instance has no '{group}.{key}' {namespace} key.")
    else:
        return fn(key)


def get_dict_item(model, group, namespace, exclude: Set[str] = frozenset()):
    """
    Fetch values from Model instance.
    """

    keys = get_dict_keys(model, group, namespace)
    prop = getattr(model, namespace)
    return {key: prop[group, key] for key in keys - exclude}


@lru_cache(512)
def get_dict_keys(model, group, namespace):
    """
    Get unique group keys for model attribute.
    """

    name = f"get_{namespace}_keys_{group}"

    fn = getattr(model, name, dict)
    keys = fn()
    extra = extra_keys(model, group, namespace)
    return {*keys, *extra}



from mundi import Region


class RegionPropertyMeta(type):
    """
    Metaclass register cached methods.
    """

    def __new__(mcs, name, bases, ns):
        return super().__new__(mcs, name, bases, ns)


class RegionProperty(metaclass=RegionPropertyMeta):
    """
    Base class for all region properties.
    """

    __slots__ = ("region",)
    region: Region

    def __init__(self, region):
        self.region = region

    def __getstate__(self):
        return self.region

    def __setstate__(self, state):
        self.region = state


def cached(fn):
    """
    Mark methods as cached.
    """
    fn.is_cached = True
    return fn


def functools_lru_backend():
    pass


def joblib_backend():
    pass


def streamlit_backend():
    pass



import operator
from collections import Sequence
from types import MappingProxyType
from typing import Iterable, Union, Type

import pandas as pd

import mundi
from mundi import Region
from pydemic.utils import extract_keys
from .model_list import ModelList
from .properties import ModelGroupClinical, ModelGroupInfo, ModelGroupResults
from .utils import map_models, prepare_data
from ..models import Model


class ModelGroup(Iterable):
    """
    A group of (usually) closely related model instances.
    """

    models: ModelList
    info: ModelGroupInfo = property(ModelGroupInfo)
    results: ModelGroupResults = property(ModelGroupResults)
    clinical: ModelGroupClinical = property(ModelGroupClinical)
    kind: Type[Model] = Model

    @property
    def dates(self):
        return self._data[0].dates

    @property
    def times(self):
        return self._data[0].times

    @property
    def names(self):
        return pd.Index(m.name for m in self)

    @classmethod
    def from_children(
        cls,
        region: Union[Region, str],
        model_cls: Type[Model],
        options=MappingProxyType({}),
        **kwargs,
    ) -> "ModelGroup":
        """
        Create a group from children of the given Region.
        """
        region: Region = mundi.region(region)
        children = region.children(**extract_keys(("deep", "type", "subtype", "which"), kwargs))

        name = kwargs.pop("name", "{region.name}")
        group = []
        if options:
            options = {mundi.region(k): v for k, v in options.items()}

        for child in children:
            opts = options.get(child, {})
            if isinstance(name, str):
                opts["name"] = name.format(region=child, **kwargs, **opts)
            group.append(model_cls(region=child, **kwargs, **opts))

        return ModelGroup(group)

    def __init__(self, models):
        self._data = list(models)
        self.models = ModelList(self, self._data)

    def __len__(self) -> int:
        return len(self.models)

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, item):
        return item in self._data

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._data[item]
        elif isinstance(item, slice):
            return ModelGroup(self._data[item])
        return prepare_data({m.name: m[item] for m in self._data})

    def __getstate__(self):
        return self._data

    def __setstate__(self, data):
        self._data = data
        self.models = ModelList(self, self._data)

    #
    # Model group API
    #
    def apply(self, method, *args, **kwargs) -> list:
        """
        Execute method on each model by name and return a list of results.
        """

        out = []
        for i, m in enumerate(self):
            try:
                fn = getattr(m, method)
            except AttributeError:
                raise AttributeError(f"{i}-th model does not have a {method} method")
            else:
                out.append(fn(*args, **kwargs))
        return out

    def apply_table(self, method, *args, **kwargs) -> pd.Series:
        """
        Similar to :method:`apply`, but returns a Series object.
        """
        data = self.apply(method, *args, **kwargs)
        return pd.Series(data, index=self.names)

    def attrs(self, attr):
        """
        Return a list of values for the given attribute on each model.

        If attr is a list, return a list of lists.
        """
        if isinstance(attr, str):
            return list(self.map(getattr, attr))
        else:
            attrs = tuple(attr)
            out = []
            for m in self:
                out.append([getattr(m, attr) for attr in attrs])
            return out

    def table(self, attr) -> Union[pd.Series, pd.DataFrame]:
        """
        Like :method:`attrs`,  but return a Series object.

        If attr is a list, return a DataFrame.
        """
        if isinstance(attr, str):
            return pd.Series(self.attrs(attr), index=self.names)
        else:
            attrs = tuple(attr)
            return pd.DataFrame(self.attrs(attrs), index=self.names, columns=attrs)

    def map(self, func, *args, **kwargs) -> Iterable:
        """
        Map function to all models and return a iterator of results.
        """
        for m in self:
            yield func(m, *args, **kwargs)

    def set_attr(self, attr, value):
        """
        Set attribute in models.
        """
        if not isinstance(value, str) and isinstance(value, Sequence):
            for m, v in zip(self, value):
                setattr(m, attr, v)
        else:
            for m in self:
                setattr(m, attr, value)

    #
    # Model API
    #
    def run(self, period):
        """
        Execute the run method of each model for the given period.
        """
        for m in self:
            m.run(period)



from typing import Union

from .covid19_class import Covid19
from .disease_class import Disease
from .disease_params import DiseaseParams
from .utils import estimate_real_cases

covid19 = Covid19("Covid-19")
DISEASE_MAP = {"covid-19": covid19, "disease": Disease("empty")}
DEFAULT = covid19


def disease(name: Union[Disease, str] = None) -> Disease:
    """
    Retrieve disease by name.

    Examples:
        >>> disease('covid-19')
        Covid19()
    """
    if name is None:
        return DEFAULT
    if isinstance(name, Disease):
        return name
    try:
        return DISEASE_MAP[name.lower()]
    except KeyError:
        diseases = ", ".join(map(repr, DISEASE_MAP.keys()))
        msg = f"invalid disease. Must be one of {diseases}, got {name!r}"
        raise ValueError(msg)


def set_default(disease):
    """
    Set the global default disease.
    """

    global DEFAULT

    if isinstance(disease, str):
        disease = globals()["disease"](disease)
    if not isinstance(disease, Disease):
        raise TypeError(f"not a disease type: {type(disease).__name__}")
    DEFAULT = disease



from pathlib import Path

import joblib
import sidekick as sk

CACHE_OPTIONS = {}


@sk.once
def user_path():
    """
    Return the user path for pydemic cache and configuration files.
    """
    path = Path("~") / ".local" / "pydemic"
    path = path.expanduser()

    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    return path


@sk.lru_cache(32)
def memory(name) -> joblib.Memory:
    """
    Return the joblib's Memory object with the given name.
    """
    if isinstance(name, joblib.Memory):
        return name

    path = user_path() / "cache" / name
    path.mkdir(parents=True, exist_ok=True)
    opts = CACHE_OPTIONS.get(name, {})
    opts.setdefault("verbose", 0)
    return joblib.Memory(path, **opts)


def set_cache_options(name, **kwargs):
    """
    Set options for the given cache.
    """
    if name in CACHE_OPTIONS and kwargs != CACHE_OPTIONS[name]:
        raise RuntimeError(f"cannot change caching options for {name}")
    CACHE_OPTIONS[name] = kwargs



import io
import re
from typing import Dict, Union

import sidekick as sk
from sidekick import placeholder as _

SPACES = re.compile(r"^\s+")
INDENT = re.compile(r"^\s+")
SECTION = re.compile(r"^\w+( \w+)*:$")
ARG = re.compile(r"^\w+(\s+\([^)]+\))?:$")


def docstring(**kwargs):
    """
    Replace values in the decorated function docstring.
    """

    def decorator(fn):
        fn.__doc__ = dedent(fn.__doc__).format(name=fn.__name__, module=fn.__module__, **kwargs)
        return fn

    return decorator


def parse_docstring(obj: Union[str, callable]) -> Dict[str, str]:
    """
    Parse a docstring in google format.

    Args:
        obj:
            Docstring or function

    Returns:
        A parsed docstring
    """
    if not isinstance(obj, str):
        doc = obj.__doc__
        if doc is None:
            raise ValueError("object does not have docstring")
    else:
        doc = obj
    lines = doc.splitlines()
    indent = len(lines[-1])
    lines = sk.pipe(
        lines,
        sk.map(_[indent:]),
        sk.partition_by(SECTION.fullmatch),
        sk.cons(("",)),
        sk.partition(2),
        dict,
    )
    return {k[0].strip(":"): "\n".join(filter(None, v)) for k, v in lines.items()}


def parse_args(args):
    """
    Parse the "Args" section of a google docstring.
    """
    out = sk.pipe(
        args.splitlines(), sk.map(str.lstrip), sk.partition_by(ARG.fullmatch), sk.partition(2), dict
    )
    return {k[0].rstrip(":"): "\n".join(v) for k, v in out.items()}


def render_docstring(parsed: Dict[str, str], indent=""):
    """
    Renders parsed docstring returned from parse_docstring.

    Args:
        parsed:
            A dictionary returned from parse_docstring()
        indent:
            An optional prefix to include in each line of the docstring.

    Returns:
        Reconstruct docstring from parsed values.
    """
    stream = io.StringIO()
    for k, v in parsed.items():
        if k:
            stream.write(k)
            stream.write(":\n")
        stream.write(v)
        stream.write("\n\n")

    out = stream.getvalue()
    if indent:
        out = "\n".join(indent + ln for ln in out.splitlines())
    return out


def dedent(doc):
    """
    Dedent according to indentation level of the last line.
    """
    lines = doc.splitlines()
    indent = SPACES.match(lines[-1]).end()
    return "\n".join(ln[indent:] for ln in lines)



import numpy as np

from pydemic.diseases import covid19
from pydemic.models import eSIR


class TestSIR:
    def test_basic_esir_api(self):
        m = eSIR(disease=covid19)
        m.run(30)
        res = m["I"]
        ok = m.data.loc[m.times[0], "infectious"] * np.exp(m.K * m.times)

        assert m.R0 == 2.74
        assert abs(m.K - m.gamma * 1.74) <= 1e-6
        assert m.iter == len(m.data) == len(m.times) == len(m.dates)
        assert np.abs(res / ok - 1).max() < 1e-4



# flake8: noqa
from mundi import Region

# noinspection PyUnresolvedReferences
from mundi_demography import age_distribution, age_pyramid, population

# noinspection PyUnresolvedReferences
from mundi_healthcare import (
    hospital_capacity,
    hospital_capacity_public,
    icu_capacity,
    icu_capacity_public,
)
from .base import RegionProperty
from .pydemic_property import PydemicProperty
from .pyplot_property import PyplotProperty


class RegionT(Region):
    plot: PyplotProperty
    pydemic: PydemicProperty


Region.pydemic = property(PydemicProperty)
Region.plot = property(PyplotProperty)



import numpy as np

from . import seir
from .base import formula

IGNORE = tuple(x for x in seir.IGNORE if x not in ("rho", "prob_symptoms"))
seair_formula = lambda ignore=(): formula("SEAIR", ignore=(*ignore, *IGNORE))

R0_from_K = seir.R0_from_K.register("SEAIR")
K = seir.K.register("SEAIR")


@seair_formula(ignore=["sigma"])
def R0(beta, gamma, prob_symptoms, rho):
    """
    R0 from model parameters.
    """
    Qs = prob_symptoms
    return beta / gamma * (Qs + (1 - Qs) * rho)


@seair_formula(ignore=["sigma"])
def beta(R0, gamma, prob_symptoms, rho):
    """
    R0 from model parameters.
    """
    Qs = prob_symptoms
    return R0 * gamma / (Qs + (1 - Qs) * rho)


#
# Initial conditions
#
time_to_seed = seir.time_to_seed.register("SEAIR")
infectious_from_cases = seir.infectious_from_cases.register("SEAIR")


@seair_formula(ignore=["rho"])
def exposed_from_cases(cases, R0, gamma, sigma, prob_symptoms):
    """
    Initializes the "exposed" component of a SEAIR model from the current
    number of cases.

    This formula assumes a perfect exponential growth.
    """
    Qs = prob_symptoms
    k = K.formula(R0, gamma, sigma)
    I0 = infectious_from_cases.formula(cases, R0, gamma, sigma)
    return I0 * (gamma + k) / (sigma * Qs)


@seair_formula(ignore=["rho"])
def asymptomatic_from_cases(cases, R0, gamma, sigma, prob_symptoms):
    """
    Initializes the "asymptomatic" component of a SEAIR model from the current
    number of cases.

    This formula assumes a perfect exponential growth.
    """
    Qs = prob_symptoms
    I0 = infectious_from_cases.formula(cases, R0, gamma, sigma)
    return I0 * (1 - Qs) / Qs


@seair_formula(ignore=["rho"])
def state_from_cases(population, cases, R0, gamma, sigma, prob_symptoms):
    """
    Initialize the 5 components of a SEAIR model from cases and population.
    """

    I = infectious_from_cases.formula(cases, R0, gamma, sigma)
    E = exposed_from_cases.formula(cases, R0, gamma, sigma, prob_symptoms)
    A = asymptomatic_from_cases.formula(cases, R0, gamma, sigma, prob_symptoms)
    S = population - I - E
    R = cases - I
    R /= prob_symptoms
    return np.array([S, E, A, I, R])



from functools import lru_cache, wraps

import numpy as np


def interpolant(x, y):
    """
    Creates a linear interpolant for the given function passed as sequences of
    x, y points.
    """
    x = np.array(x)
    y = np.array(y)

    def fn(t):
        return np.interp(t, x, y)

    return fn


def lru_safe_cache(size):
    """
    A safe LRU cache that returns a copy of the cached element to prevent
    mutations from cached values.
    """

    def decorator(func):
        cached = lru_cache(size)(func)

        @wraps(func)
        def fn(*args, **kwargs):
            return cached(*args, **kwargs).copy()

        fn.unsafe = cached
        return fn

    return decorator


def coalesce(*args, raises=False):
    """
    Return the first non-null value.

    If raises=True and no non-null value is found, raise a ValueError.

    Examples:
        >>> coalesce(None, "first", None, "second")
        'first'
    """
    for arg in args:
        if arg is not None:
            return arg
    if raises:
        raise ValueError("All values are null.")
    return None


def maybe_run(*args, **kwargs):
    """
    Return None if argument after the input function is null, otherwise
    execute fn(*args, **kwargs)

    Examples:
        >>> from math import sqrt
        >>> maybe_run(sqrt, 4)
        2
        >>> maybe_run(sqrt, None)
        None

    """
    if args[1] is None:
        return None
    args = iter(args)
    return next(args)(*args, **kwargs)



import numpy as np
import pandas as pd
import sidekick as sk
from scipy.integrate import cumtrapz
from sidekick import placeholder as _

from .hospitalization_with_delay import HospitalizationWithDelay
from ..utils import param_property, sliced
from ..mixins.info import Event


class HospitalizationWithOverflow(HospitalizationWithDelay):
    """
    Hospitals have a maximum number of ICU and regular clinical beds.
    Death rates increase when this number overflows.
    """

    hospitalization_overflow_bias = param_property(default=0.0)
    icu_capacity = param_property()
    hospital_capacity = param_property()
    icu_occupancy = param_property(default=0.75)
    hospital_occupancy = param_property(default=0.75)

    icu_surge_capacity = sk.lazy(_.icu_capacity * (1 - _.icu_occupancy))
    hospital_surge_capacity = sk.lazy(_.hospital_capacity * (1 - _.hospital_occupancy))

    def __init__(self, *args, occupancy=None, **kwargs):
        if occupancy is not None:
            kwargs.setdefault("icu_occupancy", occupancy)
            kwargs.setdefault("hospital_occupancy", occupancy)
        super().__init__(*args, **kwargs)

    def _icu_capacity(self):
        if self.region is not None:
            capacity = self.region.icu_capacity
            if np.isfinite(capacity):
                return capacity
        return self.population

    def _hospital_capacity(self):
        if self.region is not None:
            capacity = self.region.hospital_capacity
            if np.isfinite(capacity):
                return capacity
        return self.population

    #
    # Data methods
    #

    # Deaths
    def get_data_deaths(self, idx):
        return self["natural_deaths", idx] + self["overflow_deaths", idx]

    def get_data_natural_deaths(self, idx):
        """
        The number of deaths assuming healthcare system in full capacity.
        """
        return super().get_data_deaths(idx)

    def get_data_overflow_deaths(self, idx):
        """
        The number of deaths caused by overflowing the healthcare system.
        """
        return self["icu_overflow_deaths", idx] + self["hospital_overflow_deaths", idx]

    def get_data_icu_overflow_deaths(self, idx):
        """
        The number of deaths caused by overflowing ICUs.
        """
        # We just want to comput the excess deaths, so we discount the
        # contribution from natural ICUFR that is computed in natural deaths
        scale = 1 - self.ICUFR
        area = cumtrapz(self["critical_overflow"] * scale, self.times, initial=0)
        data = pd.Series(area / self.critical_period, index=self.times)
        return sliced(data, idx)

    def get_data_hospital_overflow_deaths(self, idx):
        """
        The number of deaths caused by overflowing regular hospital beds.
        """
        area = cumtrapz(self["severe_overflow"], self.times, initial=0)
        cases = area / self.severe_period
        ratio = (self.Qcr / self.Qsv) * self.hospitalization_overflow_bias
        deaths = cases * min(ratio, 1)
        data = pd.Series(deaths, index=self.times)
        return sliced(data, idx)

    def get_data_overflow_death_rate(self, idx):
        """
        Daily number of additional deaths due to overflowing the healthcare system.
        """
        return self["overflow_deaths", idx].diff().fillna(0)

    def get_data_icu_overflow_death_rate(self, idx):
        """
        Daily number of additional deaths due to overflowing the ICU capacity.
        """
        return self["icu_overflow_deaths", idx].diff().fillna(0)

    def get_data_hospital_overflow_death_rate(self, idx):
        """
        Daily number of additional deaths due to overflowing hospital capacity.
        """
        return self["hospital_overflow_deaths", idx].diff().fillna(0)

    # Severe/hospitalizations dynamics
    def get_data_severe_overflow(self, idx):
        """
        The number of severe cases that are not being treated in a hospital
        facility.
        """
        data = np.maximum(self["severe", idx] - self.hospital_surge_capacity, 0)
        return pd.Series(data, index=sliced(self.times, idx))

    def get_data_hospitalized_cases(self, idx):
        area = cumtrapz(self["hospitalized"], self.times, initial=0)
        data = pd.Series(area / self.severe_period, index=self.times)
        return sliced(data, idx)

    def get_data_hospitalized(self, idx):
        demand = self["severe", idx]
        data = np.minimum(demand, self.hospital_surge_capacity)
        return pd.Series(data, index=sliced(self.times, idx))

    # Critical/ICU dynamics
    def get_data_critical_overflow(self, idx):
        """
        The number of critical cases that are not being treated in an ICU
        facility.
        """
        data = np.maximum(self["critical", idx] - self.icu_surge_capacity, 0)
        return pd.Series(data, index=sliced(self.times, idx))

    def get_data_icu_cases(self, idx):
        area = cumtrapz(self["icu"], self.times, initial=0)
        data = pd.Series(area / self.critical_period, index=self.times)
        return sliced(data, idx)

    def get_data_icu(self, idx):
        demand = self["hospitalized", idx]
        data = np.minimum(demand, self.icu_surge_capacity)
        return pd.Series(data, index=sliced(self.times, idx))

    # Aliases
    get_data_icu_overflow = get_data_critical_overflow
    get_data_hospital_overflow = get_data_severe_overflow

    # Capacities
    def get_data_hospital_capacity(self, idx):
        return self._get_param("hospital_capacity", idx)

    def get_data_hospital_surge_capacity(self, idx):
        return self._get_param("hospital_surge_capacity", idx)

    def get_data_icu_capacity(self, idx):
        return self._get_param("icu_capacity", idx)

    def get_data_icu_surge_capacity(self, idx):
        return self._get_param("icu_surge_capacity", idx)

    def _get_param(self, name, idx, value=None):
        data = self["infectious", idx] * 0
        data.name = name
        data += getattr(self, name) if value is None else value
        return data

    #
    # Results methods
    #
    def overflow_date(self, col, value=0.0):
        """
        Get date in which column assumes a value greater than value.
        """
        for date, x in self[f"{col}:dates"].iteritems():
            if x > value:
                return date
        return None

    # TODO: convert to events
    def get_results_value_dates__icu_overflow(self):
        return self.overflow_date("critical", self.icu_surge_capacity)

    def get_results_value_dates__hospital_overflow(self):
        return self.overflow_date("severe", self.hospital_surge_capacity)

    def get_info_value_event__icu_overflow(self):
        date = self.overflow_date("critical")
        return Event.from_model(self, "icu_overflow", date)

    def get_info_value_event__hospital_overflow(self):
        date = self.overflow_date("severe")
        return Event.from_model(self, "hospital_overflow", date)



import numpy as np
import pandas as pd
from scipy import integrate
from sidekick import X

from .abstract_sir import AbstractSIR as Base
from .. import params
from ..formulas import K, R0_from_K
from ..utils import state_property, param_property, param_transform


class AbstractSEIR(Base):
    """
    Abstract base class for all SEIR-based models.
    """

    class Meta:
        model_name = "SEIR"
        data_aliases = {"E": "exposed", "exposed": None}

    # Basic epidemiological parameters
    params = params.epidemic.DEFAULT
    sigma = param_transform("incubation_period", (1 / X), (1 / X))
    incubation_period = param_property("incubation_period", default=1.0)

    # Derived expressions
    @property
    def K(self):
        return K("SEIR", gamma=self.gamma, sigma=self.sigma, R0=self.R0)

    @K.setter
    def K(self, value):
        self.R0 = R0_from_K("SEIR", gamma=self.gamma, sigma=self.sigma, K=value)

    # Simulation state
    susceptible = state_property(0)
    exposed = state_property(1)
    infectious = state_property(2)
    recovered = state_property(3)

    def _initial_state(self):
        return np.array((self.population - 1, 1, 0, 0), dtype=float)

    def _initial_infected(self):
        return super().extrapolate_cases("exposed")

    def get_data_cases(self, idx):
        # SEIR makes a distinction between "Infectious" and "Exposed". Different
        # diseases may have different clinical evolutions, but it is reasonable
        # to expect that in many situations, individuals only manifest symptoms
        # after becoming infectious.
        #
        # We therefore changed the definition of "cases" in SEIR to be simply
        # the number of elements that enter the "I" compartment.
        E = self["exposed"]
        res = integrate.cumtrapz(self.Qs * self.sigma * E, self.times, initial=0.0)
        data = pd.Series(res + self.initial_cases, index=E.index)
        return data[idx] if idx is not None else data



from pydemic import models, ModelGroup
from pydemic.clinical_models import CrudeFR


class TestModelGroup:
    def test_model_group_from_split(self):
        m = models.SIR(R0=2, gamma=1)
        m.run(10)
        grp: ModelGroup = m.split(R0=[1.25, 1.5, 2.0, 2.5], name="R0 = {R0}")

        # Basic properties
        assert len(grp) == 4
        assert grp[0].name == "R0 = 1.25"
        assert grp[0].R0 == 1.25

        # Data
        assert grp["I"].shape == (11, 4)
        assert grp["I", 5:].shape == (6, 4)
        assert grp[["I", "R"], 5:].shape == (6, 8)

        # Clinical
        assert all(isinstance(cm, CrudeFR) for cm in grp.clinical.crude_model())
        assert all(isinstance(cm, CrudeFR) for cm in grp.clinical())



from invoke import task


@task
def test(ctx, all=False, report_xml=False, verbose=False, clear_cache=False):
    if clear_cache:
        ctx.run("rm -rf .pytest_cache")
    suffix = " -vv " if verbose else ""
    if not all:
        ctx.run(f'pytest --maxfail=2 --lf -m "not slow" {suffix}', pty=True)
    if all:
        suffix += " --cov-report xml" if report_xml else ""
        ctx.run(f"pytest --cov {suffix}", pty=True)
        style(ctx)


@task
def style(ctx):
    ctx.run("black --check .")
    ctx.run("flake8 pydemic")


@task
def cov(ctx, report=True):
    ctx.run("pytest --cov --cov-report=html --cov-report=term", pty=True)


@task
def ci(ctx):
    """
    Non-configurable task that is run in continuous integration.
    """
    ctx.run("rm coverage.xml -rf")
    ctx.run("./cc-test-reporter before-build", pty=True)
    test(ctx, all=True, report_xml=True)
    ctx.run("echo 'COVERAGE.XML stats' && wc coverage.xml")
    ctx.run(
        "./cc-test-reporter after-build -r "
        "7e76aeb0890556339111bab973b18d3678572fdb2c63ea3bb388f3933066756b"
    )



from functools import lru_cache
from numbers import Number
from typing import Union, NamedTuple

import numpy as np
import pandas as pd
import sidekick as sk

import mundi
import mundi_demography as mdm
from pydemic import db
from pydemic.params import get_param
from pydemic.utils import slugify

QualDataT = Union[pd.DataFrame, "Dataset"]
QualValueT = Union[pd.Series, Number, "Dataset"]
DB_PATH = db.DATABASES / "diseases"


def dunder(name, reversed=False):
    rname = ("__" + name[3:]) if reversed else ("__r" + name[2:])

    def op(self: "Dataset", other):
        if isinstance(other, Dataset):
            other = other.data

        fn = getattr(self.data, name, None)
        if fn is not None:
            value = fn(other)
            if value is not NotImplemented:
                return Dataset(value)

        fn = getattr(other, rname, None)
        if fn is not None:
            value = fn(self.data)
            if value is not NotImplemented:
                return Dataset(value)

        return NotImplemented

    op.__name__ = op.__qualname__ = name
    return op


class Dataset(NamedTuple):
    """
    A qualified dataset. Includes raw data in the .data attribute and meta
    information
    """

    data: Union[pd.DataFrame, pd.Series, np.ndarray, Number]
    source: str = ""
    notes: str = ""

    __add__ = dunder("__add__")
    __sub__ = dunder("__sub__")
    __mul__ = dunder("__mul__")
    __truediv__ = dunder("__truediv__")
    __floordiv__ = dunder("__floordiv__")
    __pow__ = dunder("__pow__")
    __mod__ = dunder("__mod__")

    # Reversed methods
    __radd__ = dunder("__radd__", reversed=True)
    __rsub__ = dunder("__rsub__", reversed=True)
    __rmul__ = dunder("__rmul__", reversed=True)
    __rtruediv__ = dunder("__rtruediv__", reversed=True)
    __rfloordiv__ = dunder("__rfloordiv__", reversed=True)
    __rpow__ = dunder("__rpow__", reversed=True)
    __rmod__ = dunder("__rmod__", reversed=True)


def normalize_source(source):
    """
    Slugify and remove unnecessary suffixes from source reference.

    This is useful to transform a textual description of the source work to
    a valid filename in which it is installed.

    Examples:
        >>> normalize_source("Verity, et. al.")
        "verity"
    """
    return slugify(source, suffixes=("et-al",))


def lazy_stored_string(path):
    """
    A lazy string-like object that loads information from a file.
    """
    return sk.deferred(lazy_description, path)


def age_adjusted_average(ages, value):
    """
    Given an age distribution and a table with age-adjusted values, compute the
    mean value for population.

    This method automatically aligns ages and value if both tables are not
    aligned.
    """
    if isinstance(ages, str):
        ages = mdm.age_distribution(ages)
    population = ages.values.sum()
    if set(ages.index).issuperset(value.index):
        value = value.reindex(ages.index, method="ffill")
    else:
        raise NotImplementedError
    return (value * ages / population).sum()


@lru_cache(1)
def world_age_distribution():
    """
    World age distribution computed by summing
    """
    countries = mundi.countries()
    return countries.mundi["age_distribution"].sum(0)


@lru_cache(32)
def lazy_description(path):
    """
    Lazily load string from path.
    """
    with open(DB_PATH / path) as fd:
        return fd.read()


@lru_cache(32)
def read_table(path: str, key=None):
    """
    Read dataframe from path.
    """
    return db.read_table(path, key)


def set_age_distribution_default(dic, value=None, drop=False):
    """
    Set the ages_distribution key of dictionary to the given value or to the
    World's age distribution.
    """

    ages = dic.pop("age_distribution", None)

    if ages is None:
        ages = world_age_distribution() if value is None else value
        if isinstance(ages, str):
            ages = mdm.age_distribution(value)
        elif not isinstance(ages, (pd.Series, pd.DataFrame)):
            ages = get_param("age_distribution", value)

    if not drop:
        dic["age_distribution"] = ages
    return ages


def estimate_real_cases(curves: pd.DataFrame, params=None, method="CFR") -> pd.DataFrame:
    """
    Estimate the real number of cases from the cases and deaths curves.

    Returns:
        A new dataframe with the corrected "cases" and "deaths" columns.
    """

    if params is None:
        from . import disease

        params = disease().params()

    data = curves[["cases", "deaths"]]

    if method == "CFR":
        daily = data.diff().dropna()
        cases, deaths = daily[(daily != 0).all(axis=1)].values.T
        weights = np.log(deaths)
        empirical_CFR = (weights * deaths / cases).sum() / weights.sum()
        if np.isnan(empirical_CFR):
            empirical_CFR = 0.0

        try:
            CFR = params.CFR
        except AttributeError:
            CFR = params.disease_params.CFR

        data["cases"] *= max(empirical_CFR, CFR) / CFR
        return data
    else:
        raise ValueError(f"Invalid estimate method: {method!r}")



import inspect
from collections import defaultdict
from functools import wraps
from numbers import Real

import numpy as np

from ..params import get_param, ParamLike

# Function registry
FUNCTIONS_R0_FROM_K = {}
FUNCTIONS_K = {}
FUNCTIONS_R0 = {}
FUNCTIONS_BETA = {}
FUNCTIONS_STATE_FROM_CASES = {}
FUNCTIONS = defaultdict(
    dict,
    {
        "R0_from_K": FUNCTIONS_R0_FROM_K,
        "K": FUNCTIONS_K,
        "R0": FUNCTIONS_R0,
        "beta": FUNCTIONS_BETA,
        "state_from_cases": FUNCTIONS_STATE_FROM_CASES,
    },
)

# Constants
LN_2 = np.log(2)

# Aliases and alternate forms for the "formula" decorator.
ALTERNATE_FORMS = {
    # Inverted relations
    "infectious_period": ("gamma", lambda x: 1 / x),
    "incubation_period": ("sigma", lambda x: 1 / x),
    "doubling_time": ("K", lambda k: float("inf") if k == 0 else LN_2 / k),
    # Aliases
    "Qs": ("prob_symptoms", lambda x: x),
}


def formula(model=None, ignore=(), options=(), invalid="raise", positional=0, formula_name=None):
    """
    Decorator that register formula functions.

    Formula functions receive a simple function that computes a mathematical
    formula from numeric keyword arguments and return a more flexible interface
    which can accept arguments from a namespace, understand alternative
    representations and aliases for common arguments, etc.

    Args:
        model:
            Name of the epidemiological model the formula applies to. If a
            sequence is passed, register function for several different models.
        ignore:
            Sequence of arguments to ignore. This is useful if one wants to
            provide a consistent interface for different models. This way, a
            simple model can simply ignore parameters from a more sophisticated
            one (e.g.. SIR can ignore the incubation_period of a SEIR model).
        options:
            Additional parameters that are simply passed to the decorated
            function without any special handling.
        positional:
            The number of required positional arguments the function uses.
        invalid ('ignore', 'pass', 'raise'):
            Strategy to use when encounter invalid parameters: 'ignore', simply
            suppress them, 'pass', pass them to the decorated function and
            'raise' (default) raises an error.
        formula_name:
            If given, register function under a different name than the declared
            function name.
    """

    def decorator(fn):
        if invalid not in ("ignore", "pass", "raise"):
            raise ValueError(f'bad "invalid" strategy: {invalid!r}')

        signature = inspect.signature(fn)
        arguments = set(list(signature.parameters)[positional:]).difference(options)
        alternatives = {x for x, (k, _) in ALTERNATE_FORMS.items() if k in arguments}
        fn_name = formula_name or fn.__name__

        # Remove **starred_kwargs from arguments
        for arg in signature.parameters.values():
            if arg.kind == arg.VAR_KEYWORD:
                arguments.discard(arg.name)
        n_args = len(arguments)

        # Feed ignore set with the list of alternatives
        ignore_set = set(ignore)
        options_set = set(options)

        for k in ignore_set:
            try:
                k, _ = ALTERNATE_FORMS[k]
                ignore_set.add(k)
            except KeyError:
                continue

        @wraps(fn)
        def decorated(*args_, **kwargs):
            params: ParamLike = None

            options = {}
            args = args_[:positional]
            if len(args_) == positional:
                params = None
            elif len(args_) == positional + 1:
                params = args_[positional]
            else:
                n = positional
                m = len(args_)
                TypeError(f"Expect {n} or {n + 1} positional arguments, got {m}")

            args = args[:positional]

            # Remove all ignored parameters
            if ignore_set:
                for k in list(kwargs):
                    if k in ignore_set:
                        del kwargs[k]

            # Extract options
            for k in options_set.intersection(kwargs):
                options[k] = kwargs.pop(k)

            # Normalize all alternative values passed as keyword arguments.
            for k in alternatives.intersection(kwargs):
                v = kwargs.pop(k)
                k, transform = ALTERNATE_FORMS[k]
                kwargs[k] = transform(v)

            # Check for invalid keyword arguments
            invalid_set = set(kwargs) - arguments
            if invalid_set and invalid == "ignore":
                for k in invalid:
                    del kwargs[k]
            elif invalid_set and invalid == "raise":
                raise TypeError(f"invalid argument: {invalid_set.pop()}")
            elif invalid_set:
                options.update({k: kwargs.pop(k) for k in invalid_set})

            if len(kwargs) == n_args:
                return fn(*args, **kwargs, **options)

            # If we do not have enough arguments, try the positional value
            missing = arguments - kwargs.keys()
            if params is None:
                raise TypeError(f"missing required argument: {missing.pop()}")

            for k in missing:
                try:
                    kwargs[k] = get_param(k, params)
                except ValueError:
                    pass

            # Try alternate forms if not all keyword arguments were completed.
            if len(kwargs) < n_args:
                for argname in alternatives:
                    k, transform = ALTERNATE_FORMS[argname]
                    if k not in kwargs:
                        kwargs[k] = transform(get_param(argname, params))

            return fn(*args, **kwargs, **options)

        def register(model):
            """
            Register function to model or list of models.

            Return the decorated function.
            """

            lst = (model,) if isinstance(model, str) else tuple(model)
            decorated.models = frozenset([*lst, *decorated.models])
            for m in lst:
                FUNCTIONS[fn.__name__][m] = decorated
            return decorated

        # Save attributes to the decorated function
        decorated.formula = fn
        decorated.arguments = frozenset(arguments)
        decorated.alternate_arguments = frozenset(alternatives)
        decorated.register = register

        # Register formula
        model_lst = (model,) if isinstance(model, str) else tuple(model or ())
        for m in model_lst:
            FUNCTIONS[fn_name][m] = decorated
        decorated.models = frozenset(model_lst)

        return decorated

    return decorator


#
# Generic formulas
#
def R0(model, params=None, **kwargs) -> float:
    """
    Compute R0 for the given model using epidemiological parameters.

    Args:
        model ({'SIR', 'SEIR', 'SEAIR', etc}):
            A string naming the desired model.
        params:
            A params object as described in :func:`get_param`.

    Keyword Args:
        Extra epidemiological parameters that may be used in the formula. This
        overrides the values passed in params. It is safe to pass non-used
        parameters, and they will be simply ignored.

    Examples:
        >>> R0("SIR", beta=1.0, gamma=0.5)
        2.0

    See Also:
        :func:`sir.R0`
        :func:`seir.R0`
        :func:`seair.R0`

    Returns:
        The value of R0 for that model.
    """
    return FUNCTIONS_R0[model](params, **kwargs)


def R0_from_K(model, params=None, **kwargs) -> float:
    """
    Compute R0 for the given model using the exponential growth factor K.

    Args:
        model ({'SIR', 'SEIR', 'SEAIR', etc}):
            A string naming the desired model.
        params:
            A params object as described in :func:`get_param`.

    Keyword Args:
        Extra epidemiological parameters that may be used in the formula. This
        overrides the values passed in params. It is safe to pass non-used
        parameters, and they will be simply ignored.

    Examples:
        >>> R0_from_K("SIR", K=0.5, gamma=0.5)
        2.0

    See Also:
        :func:`sir.R0_from_K`
        :func:`seir.R0_from_K`
        :func:`seair.R0_from_K`

    Returns:
        The value of R0 for that model.
    """
    return FUNCTIONS_R0_FROM_K[model](params, **kwargs)


def beta(model, params=None, **kwargs) -> float:
    """
    Compute beta from for the given model using epidemiological parameters.

    Args:
        model ({'SIR', 'SEIR', 'SEAIR', etc}):
            A string naming the desired model.
        params:
            A params object as described in :func:`get_param`.

    Keyword Args:
        Extra epidemiological parameters that may be used in the formula. This
        overrides the values passed in params. It is safe to pass non-used
        parameters, and they will be simply ignored.

    Examples:
        >>> beta("SIR", R0=2.0, gamma=0.5)
        1.0

    See Also:
        :func:`sir.beta`
        :func:`seir.beta`
        :func:`seair.beta`

    Returns:
        The value of beta for that model.
    """
    return FUNCTIONS_BETA[model](params, **kwargs)


def K(model, params=None, **kwargs) -> float:
    """
    Compute the exponential growth factor K for the given model.

    Args:
        model ({'SIR', 'SEIR', 'SEAIR', etc}):
            A string naming the desired model.
        params:
            A params object as described in :func:`get_param`.

    Keyword Args:
        Extra epidemiological parameters that may be used in the formula. This
        overrides the values passed in params. It is safe to pass non-used
        parameters, and they will be simply ignored.

    Examples:
        >>> K("SIR", R0=2.0, gamma=0.5)
        0.5

    Returns:
        The value of K for that model.
    """
    return FUNCTIONS_K[model](params, **kwargs)


def doubling_time(model, params=None, **kwargs) -> Real:
    """
    Compute the doubling time for the given model. Negative results correspond
    to the halving time of decaying processes.

    Args:
        model ({'SIR', 'SEIR', 'SEAIR', etc}):
            A string naming the desired model.
        params:
            A params object as described in :func:`get_param`.

    Keyword Args:
        Extra epidemiological parameters that may be used in the formula. This
        overrides the values passed in params. It is safe to pass non-used
        parameters, and they will be simply ignored.

    Returns:
        The doubling (or halving) time for the process.
    """
    k = FUNCTIONS_K[model](params, **kwargs)
    if k == 0:
        return float("inf")
    else:
        return LN_2 / k


def initial_state(model, cases, params=None, **kwargs) -> np.ndarray:
    """
    Compute the initial state for model using the given number of cases.

    Args:
       model ({'SIR', 'SEIR', 'SEAIR', etc}):
           A string naming the desired model.
       params:
           A params object as described in :func:`get_param`.

    Keyword Args:
       Extra epidemiological parameters that may be used in the formula. This
       overrides the values passed in params. It is safe to pass non-used
       parameters, and they will be simply ignored.

    Returns:
       An array with the estimated initial state.
    """
    return FUNCTIONS_STATE_FROM_CASES[model](params, cases=cases, **kwargs)


def get_function(model, name):
    """
    Return registered function.
    """
    return FUNCTIONS[name][model]



from .abstract_seir import AbstractSEIR


class eSEIR(AbstractSEIR):
    """
    A simple SEIR model linearized around the DFE.
    """

    def run_to_fill(self, data, ts):
        raise NotImplementedError



from pydemic.models import SEAIR
from pydemic.diseases import covid19


class TestSEAIR:
    def test_seair_initialization(self):
        # Probability of developing symptoms
        assert SEAIR(disease="disease").prob_symptoms == 1.0
        assert SEAIR(disease=covid19).prob_symptoms == covid19.prob_symptoms()
        assert SEAIR(prob_symptoms=0.5).prob_symptoms == 0.5
        assert SEAIR(Qs=0.5).prob_symptoms == 0.5



from gettext import gettext as _

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pydemic.types import ValueStd
from . import helpers
from .. import fitting as fit
from ..diseases import disease as get_disease
from ..utils import trim_weeks, weekday_name, accumulate_weekly


def cases_and_deaths(
    data: pd.DataFrame,
    dates: bool = False,
    ax: plt.Axes = None,
    smooth: bool = True,
    cases: str = "cases",
    deaths: str = "deaths",
    tight_layout=False,
    **kwargs,
) -> plt.Axes:
    """
    A simple chart showing observed new cases cases as vertical bars and
    a smoothed out prediction of this curve.

    Args:
        data:
            A dataframe with ["cases", "deaths"] columns.
        dates:
            If True, show dates instead of days in the x-axis.
        ax:
            An explicit matplotlib axes.
        smooth:
            If True, superimpose a plot of a smoothed-out version of the cases
            curve.
        cases:
        deaths:
            Name of the cases/deaths columns in the dataframe.
    """

    if not dates:
        data = data.reset_index(drop=True)

    # Smoothed data
    col_names = {cases: _("Cases"), deaths: _("Deaths")}
    if smooth:
        from pydemic import fitting as fit

        smooth = pd.DataFrame(
            {
                _("{} (smooth)").format(col_names[cases]): fit.smoothed_diff(data[cases]),
                _("{} (smooth)").format(col_names[deaths]): fit.smoothed_diff(data[deaths]),
            },
            index=data.index,
        )
        ax = smooth.plot(legend=False, lw=2, ax=ax)

    # Prepare cases dataframe and plot it
    kwargs.setdefault("alpha", 0.5)
    new_cases = data.diff().fillna(0)
    new_cases = new_cases.rename(col_names, axis=1)

    if "ylim" not in kwargs:
        deaths = new_cases.iloc[:, 1]
        exp = np.log10(deaths[deaths > 0]).mean()
        exp = min(10, int(exp / 2))
        kwargs["ylim"] = (10 ** exp, None)
    ax: plt.Axes = new_cases.plot.bar(width=1.0, ax=ax, **kwargs)

    # Fix xticks
    periods = 7 if dates else 10
    xticks = ax.get_xticks()
    labels = ax.get_xticklabels()
    ax.set_xticks(xticks[::periods])
    ax.set_xticklabels(labels[::periods])
    ax.tick_params("x", rotation=0)
    ax.set_ylim(1, None)
    if tight_layout:
        fig = ax.get_figure()
        fig.tight_layout()
    return ax


def weekday_rates(
    data, *, trend=False, trend_style="k--", xtick_rotation=45, translate=None, **kwargs
):
    """
    Plot fractions of weekly values accumulated at each week day.

    Args:
        data:
            Data frame with time series data
        trend:
            If given, show a trend line with the expected value that should be
            encountered in each week day.
            * False (default): omit trend line
            * {True, 'cte'}: show a horizontal trend as if each week day should
              contribute the same amount to the weekly total
            * 'exp': Assumes an exponential growth;
        trend_style:
            If given, overrides the default style ('k--') for the trend line.
        xtick_rotation:
            Control rotation of x-ticks.

    Keyword Args:
        Additional keyword arguments are passed to the plot method of a
        weekly rates dataframe.
    """

    kwargs.setdefault("width", 0.75)

    data = trim_weeks(data)
    rate = fit.weekday_rate(data)
    std = rate[[col + "_std" for col in data.columns]]
    values = rate[data.columns]
    values.index = (weekday_name(dt, translate=translate) for dt in values.index)
    ax = values.plot(kind="bar", yerr=std.values.T, **kwargs)
    ax.tick_params("x", rotation=xtick_rotation)

    if trend == "exp":
        by_week = accumulate_weekly(data)
        growth, growth_std = ValueStd.mean(fit.growth_factors(by_week))
        daily_growth = growth ** (1 / 7)

        X = np.arange(-0.5, 7.5)
        trend = daily_growth ** X
        trend /= trend.sum()
        ax.plot(X)

    elif trend == "cte" or trend is True:
        helpers.mark_x(1 / 7, trend_style)

    elif trend is not False:
        raise ValueError("invalid trending method")

    return ax


def plot_weekday_rate_for_region(region, disease=None, **kwargs):
    """
    Like plot_weekly_rate, but cache results per-region.
    """

    disease = get_disease(disease)
    curves = disease.epidemic_curve(region)
    values = np.diff(curves, prepend=0, axis=0)

    data = pd.DataFrame(values, index=curves.index, columns=curves.columns)
    data = trim_weeks(data)

    return weekday_rates(data, **kwargs)



import numpy as np
import pandas as pd

from .epidemic_curves import infectious_curve
from .. import formulas
from ..formulas import formula

CURVE_OPTIONS = ("dt", "Rt_smooth", "smooth_days", "ret_Rt")


@formula("SIR", positional=1, options=CURVE_OPTIONS, invalid="pass")
def sir_curves_Rt(
    cases, gamma, R0, population, dt=1, Rt_smooth=0.2, smooth_days=3, ret_Rt=False, **kwargs
):
    """
    Infer SIR curves from experimental data.
    """
    N = len(cases)

    infectious = infectious_curve(cases, gamma, **kwargs)
    data = np.zeros((N, 3), dtype=float)
    data[0, 0] = population
    data[:, 1] = infectious

    Rt = [1.0] if R0 is None else [R0]
    beta = formulas.beta("SIR", R0=Rt[0], gamma=gamma)

    for i, row in enumerate(data[:-1], start=1):
        S, I, R = row

        if Rt_smooth != 1:
            e = 1e-9
            xs = infectious[i - min(i, smooth_days) : i + 1]
            growth = np.diff(np.log(xs + e)).mean()
            beta_new = (growth + gamma) * (population / S)
            beta = (1 - Rt_smooth) * beta_new + Rt_smooth * beta

        Rt.append(beta / gamma)
        data[i, 2] = R + gamma * I * dt
        data[i, 0] = population - data[i, 1:].sum()

    columns = ("susceptible", "infectious", "recovered")
    df = pd.DataFrame(data, columns=columns, index=cases.index)
    if ret_Rt:
        df["Rt"] = Rt
    return df


@formula("SEIR", positional=1, options=CURVE_OPTIONS, invalid="pass")
def seir_curves_Rt(
    cases, gamma, sigma, R0, population, dt=1, Rt_smooth=0.2, smooth_days=3, ret_Rt=False, **kwargs
):
    """
    Infer SEIR curves from experimental data.
    """
    N = len(cases)

    infectious = infectious_curve(cases, gamma=gamma, **kwargs)
    data = np.zeros((N, 4), dtype=float)
    data[0, 0] = population
    data[:, 1] = infectious * (sigma / gamma)
    data[:, 2] = infectious

    Rt = [1.0] if R0 is None else [R0]
    beta = formulas.beta("SEIR", R0=Rt[0], gamma=gamma, sigma=sigma)

    for i, row in enumerate(data[:-1], start=1):
        S, E, I, R = row

        if Rt_smooth != 1:
            e = 1e-9
            xs = infectious[i - min(i, smooth_days) : i + 1]
            growth = np.diff(np.log(xs + e)).mean()
            beta_new = ((growth + sigma) * E / (I + e)) * (population / S)
            beta = (1 - Rt_smooth) * beta_new + Rt_smooth * beta

        Rt.append(beta / gamma)

        if R0 is not None:
            force = beta * I * (S / population)
            data[i, 1] = E + (force - sigma * E) * dt

        data[i, 3] = R + gamma * I * dt
        data[i, 0] = population - data[i, 1:].sum()

    columns = ("susceptible", "exposed", "infectious", "recovered")
    df = pd.DataFrame(data, columns=columns, index=cases.index)
    if ret_Rt:
        df["Rt"] = Rt
    return df


@formula("SEAIR", positional=1, options=CURVE_OPTIONS, invalid="pass")
def seair_curves_Rt(
    cases,
    gamma,
    sigma,
    prob_symptoms,
    rho,
    R0,
    population,
    dt=1,
    Rt_smooth=1.0,
    smooth_days=3,
    ret_Rt=False,
    **kwargs
):
    """
    Infer SEAIR curves from experimental data.
    """

    N = len(cases)
    infectious = infectious_curve(cases, gamma=gamma, **kwargs)

    # Allocate data and initialize with values
    data = np.zeros((N, 5), dtype=float)
    S, E, A, I, R = range(5)
    data[0, S] = population
    data[:, E] = infectious * (sigma / gamma / prob_symptoms)
    data[:, A] = infectious * (1 - prob_symptoms) / prob_symptoms
    data[:, I] = infectious

    Rt = [1.0] if R0 is None else [R0]
    formula_args = dict(gamma=gamma, sigma=sigma, rho=rho, prob_symptoms=prob_symptoms)
    beta = formulas.beta("SEAIR", R0=Rt[0], **formula_args)

    for i, row in enumerate(data[:-1], start=1):
        S, E, A, I, R = row

        if Rt_smooth != 1:
            e = 1e-9
            xs = infectious[i - min(i, smooth_days) : i + 1]
            growth = np.diff(np.log(xs + e)).mean()
            beta_new = ((growth + sigma) * E / (I + rho * A + e)) * (population / S)
            beta = (1 - Rt_smooth) * beta_new + Rt_smooth * beta

        Rt.append(formulas.R0("SEAIR", beta=beta, **formula_args))

        if R0 is not None:
            force = beta * (S / population) * (I + rho * A)
            data[i, 1] = E + (force - sigma * E) * dt

        data[i, 4] = R + gamma * (I + A) * dt
        data[i, 0] = population - data[i, 1:].sum()

    columns = ("susceptible", "exposed", "asymptomatic", "infectious", "recovered")
    df = pd.DataFrame(data, columns=columns, index=cases.index)
    if ret_Rt:
        df["Rt"] = Rt
    return df



import numpy as np

from .abstract_sir import AbstractSIR
from .ode_model import ODEModel


class SIR(ODEModel, AbstractSIR):
    """
    A simple SIR model linearized around the DFE.
    """

    def diff(self, x, t):
        s, i, r = x
        n = s + i + r
        beta = self.beta
        gamma = self.gamma
        return np.array([-beta * s * (i / n), +beta * s * (i / n) - gamma * i, +gamma * i])


if __name__ == "__main__":
    import click

    @click.command()
    @click.option("--R0", "-r", default=2.74, help="Reproduction number")
    @click.option("--duration", "-t", default=90, type=int, help="Duration")
    @click.option("--linear", "-l", is_flag=True, help="Use log-scale?")
    def cli(duration, r0, linear):
        m = SIR(R0=r0)
        m.run(duration)
        m.plot(logy=not linear, show=True)

    cli()



from pytest import approx as _approx

import mundi
from pydemic.diseases import disease as get_disease
from pydemic.models import SIR


class TestResults:
    def test_results(self):
        approx = lambda x: _approx(x, rel=0.005)
        m = SIR(disease="covid-19", region="BR")
        m.set_ic(cases=1e6)
        m.run(60)

        # Data
        assert m.results["data.attack_rate"] == approx(0.919)
        assert m.results["data.cases"] == approx(194236068)
        assert m.results["data.infectious"] == approx(4045)
        assert m.results["data.recovered"] == approx(194632955)
        assert m.results["data.susceptible"] == approx(17118720)
        assert m.results["data"] == approx(
            {
                "attack_rate": 0.919,
                "cases": 194236068,
                "infectious": 4045,
                "recovered": 194632955,
                "susceptible": 17118720,
            }
        )

        # Parameters
        assert m.results["params.R0"] == m.R0
        assert m.results["params.gamma"] == m.gamma
        assert m.results["params.infectious_period"] == m.infectious_period
        assert m.results["params"] == approx({"R0": 2.74, "infectious_period": 3.47})

        # Dates
        dates = m.results["dates"]
        assert dates["start"] < dates["peak"] < dates["end"]

    def test_can_store_information_in_results(self):
        m = SIR(disease="covid-19", region="BR")
        m.run(10)

        expect = {"bar": 42, "spam": "eggs"}
        m.results["foo.bar"] = 42
        m.results["foo.spam"] = "eggs"
        assert m.results["foo"] == expect
        assert "foo" in m.results.keys()

        assert "foo" in m.results.to_dict()
        assert "foo.bar" in m.results.to_dict(flat=True)

        m.run(10)
        assert m.results.get("foo") == {}


class TestInfo:
    def test_info(self):
        approx = lambda x: _approx(x, rel=0.005)
        disease = get_disease("covid-19")
        m = SIR(disease="covid-19", region="BR")
        m.set_ic(cases=1e6)
        m.run(60)

        # Disease
        infectious_period = disease.infectious_period(region="BR")
        assert m.info["disease.CFR"] == approx(disease.CFR(region="BR"))
        assert m.info["disease.IFR"] == approx(disease.IFR(region="BR"))
        assert m.info["disease.infectious_period"] == approx(infectious_period)
        assert set(m.info["disease"]) == {
            "R0",
            "case_fatality_ratio",
            "critical_delay",
            "critical_period",
            "death_delay",
            "hospital_fatality_ratio",
            "hospitalization_overflow_bias",
            "hospitalization_period",
            "hospitalization_table",
            "icu_fatality_ratio",
            "icu_period",
            "incubation_period",
            "infection_fatality_ratio",
            "infectious_period",
            "mortality_table",
            "prob_aggravate_to_icu",
            "prob_critical",
            "prob_severe",
            "prob_symptoms",
            "rho",
            "severe_delay",
            "severe_period",
            "symptom_delay",
        }

        # Region
        br = mundi.region("BR")
        keys = {
            "population",
            "age_distribution",
            "age_pyramid",
            "hospital_capacity",
            "icu_capacity",
        }

        assert m.info["region.population"] == br.population
        assert all(m.info["region.age_distribution"] == br.age_distribution)
        assert all(m.info["region.age_pyramid"] == br.age_pyramid)
        assert set(m.info["region"]) == keys



# flake8: noqa
from .epidemic_curves import infectious_curve, sir_curves, seir_curves, seair_curves, epidemic_curve
from .exponential_growth import (
    growth_factor,
    growth_factors,
    exponential_extrapolation,
    R0_from_cases,
    growth_ratio_from_cases,
    growth_factor_from_cases,
)
from .smoothing import smoothed_diff
from .time_dependent_growth import time_dependent_K, smoothed_diff, naive_holt_smoothing
from .seasonal import weekday_rate
from . import Rt
from . import R0
from . import K
from .Rt import estimate_Rt
from .K import estimate_K, estimate_Kt
from .R0 import estimate_R0
from .utils import smooth, diff, cases



import numpy as np
import pandas as pd
import statsmodels.api as sm

from .smoothing import smoothed_diff


def time_dependent_K(cases: pd.Series, method="smoothing", **kwargs):
    """
    Compute a time-dependent exponential growth factor from the given times
    series of cases, deaths or some strongly correlated proxy.

    Args:
        cases:
            Time-series of cases.

    Returns:
        A DataFrame with ["value", "low", "high"] columns. The  main inference
        is in the "value" interval. If the chosen method does not compute
        confidence intervals, low and high will contain only NaNs.
    """
    compute = TimeDependentK(cases)
    return compute(method, **kwargs)


#
# Class-based facade to smoothing methods
#
class TimeDependentK:
    """
    A class that implements inference methods for time-dependent K as instance
    methods in the class.
    """

    def __init__(self, cases):
        self.cases = cases
        self.dates = cases.index
        self.data = cases.values

    def __call__(self, method, **kwargs):
        fn = getattr(self, "method_" + method)
        out = fn(**kwargs)
        return pd.Series(out, index=self.dates)

    def method_smoothing(self, **kwargs):
        res = naive_holt_smoothing(self.data, **kwargs)
        return res.fittedvalues


#
# Private implementations of smoothing methods.
# Those methods do not pre-process input or post-process outputs and generally
# work with raw numpy arrays.
#
def naive_holt_smoothing(ys, max_daily_growth=2, **kwargs):
    """
    A double-smoothing method.
    """

    max_factor = np.log(max_daily_growth)
    deltas = smoothed_diff(ys, **kwargs)
    dirty_ks = np.minimum(np.diff(np.log(deltas), prepend=0), max_factor)

    smoothing_level = kwargs.setdefault("smoothing_level", 0.1)
    holt = sm.tsa.Holt(dirty_ks, damped=True)
    return holt.fit(smoothing_level=smoothing_level)



import os
from pathlib import Path

import pandas as pd

DATABASES = Path(__file__).absolute().parent / "databases"
EXT_PRIORITY = {
    ".pkl.bz2": pd.read_pickle,
    ".pkl.gz": pd.read_pickle,
    ".pkl": pd.read_pickle,
    ".csv": lambda p: pd.read_csv(p, index_col=0, parse_dates=True),
}
EXT_PRIORITY_WITH_KEYS = {
    ".hdf5": pd.read_hdf,
    ".sqlite": (lambda p, tb: pd.read_sql(f"SELECT * FROM {tb};", p)),
}


def read_table(ref: str, key=None):
    """
    Read data in a table file.
    """
    *dirs, name = ref.split("/")
    n = len(name)
    basedir = DATABASES
    for directory in dirs:
        basedir = basedir / directory

    paths = {f[n:]: basedir / f for f in os.listdir(basedir) if f.startswith(name)}
    if key:
        priorities = EXT_PRIORITY_WITH_KEYS
        reader = lambda f, p: f(p, key)
    else:
        priorities = EXT_PRIORITY
        reader = lambda f, p: f(p)

    for ext, fn in priorities.items():
        try:
            path = paths[ext]
        except KeyError:
            continue
        else:
            return reader(fn, path)
    raise ValueError(f"no data source found for {ref!r}")



import logging
import os

logging.basicConfig(format="%(levelname)s: %(message)s")
log = logging.getLogger("pydemic")

if os.environ.get("DEBUG", "").lower() in ("1", "true", "on"):
    log.setLevel(logging.DEBUG)
    log.info("Debug enabled")



import functools
import sys
import time
from functools import wraps
from types import MappingProxyType
from typing import NamedTuple

import numpy as np
import sidekick as sk

from .base import Property


def method_as_function(method, wrapper=None):
    """
    Take a method from a property object and transform it in a independent
    function that takes the owner object as first input.
    """

    if wrapper is None:
        try:
            mod = sys.modules[method.__module__]
            path = method.__qualname__.partition(".")[0]
            wrapper = getattr(mod, path)
        except AttributeError:
            pass
        else:
            wrapper = lambda x: Property(x)

    @wraps(method)
    def fn(_object, *args, **kwargs):
        return method(wrapper(_object), *args, **kwargs)

    return fn


def function_as_method(fn):
    """
    Convert function that receives value to a method of a property accessor
    for that value.
    """

    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        return fn(self._object, *args, **kwargs)

    return wrapped


@sk.fn.curry(1)
def cached(fn, ttl=None, small=None):
    """
    Tips toward a caching mechanism.

    Currently, it is a no-op, but we plan to make an actual caching
    implementation in the future.
    """
    return fn


@sk.fn.curry(1)
def cached_small_value(fn, ttl=False, maxsize=None):
    """
    Small cached values are stored in memory with an optional ttl.
    """

    cache = {}
    clean_size = max(maxsize // 2, 1)

    @wraps(fn)
    def decorated(self, *args, **kwargs):
        key = Args.from_signature(*args, **kwargs)

        try:
            result = cache[key]
        except KeyError:
            value = get_cached_small_value(self, fn, *args, **kwargs)
            result = value, (time.time() + ttl if ttl else None)

        if maxsize and len(cache) >= maxsize:
            for k in list(sk.islice(cache, clean_size)):
                del cache[k]
        cache[key] = result
        return result[0]

    return decorated


def get_cached_small_value(*args, **kwargs):
    """
    Non-decorated base implementation of a method that fetches a property from
    a method of a class.

    Method receives
    """
    prop, method, *args = args
    if isinstance(method, str):
        fn = getattr(prop, method)
    else:
        fn = method.__get__(prop)
    return fn(*args, **kwargs)


class Args(NamedTuple):
    args: tuple
    kwargs: MappingProxyType
    hash: int

    @classmethod
    def from_signature(cls, *args, **kwargs):
        hash_key = hash(tuple(iter_hash(args, kwargs)))
        return Args(args, MappingProxyType(kwargs), hash_key)

    def __hash__(self):
        return self.hash

    def __iter__(self):
        yield self.args
        yield self.kwargs


@functools.singledispatch
def value_hash(x):
    return hash(x)


@value_hash.register(np.ndarray)
def _(x):
    return hash(x.tostring())


@value_hash.register(list)
def _hash_sequence(x):
    return hash(tuple(map(value_hash, x)))


def iter_hash(args, kwargs, value_hash=value_hash):
    for i, arg in enumerate(args):
        yield (i, value_hash(arg))
    for k, v in sorted(kwargs.items()):
        yield (k, value_hash(v))



from functools import total_ordering

from ..packages import sk

NOT_GIVEN = object()


@total_ordering
class state_property(property):
    """
    A property that exposes a state component as rw.
    """

    name: str
    index: int
    ro: bool

    def __init__(self, i, ro=False, name=None):
        def fget(self):
            return self.state[i]

        def fset(self, value):
            self.state[i] = value

        args = (fget,) if ro else (fget, fset)
        self.index = i
        self.name = name
        self.ro = ro
        super().__init__(*args)

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

    def __gt__(self, other):
        if isinstance(other, state_property):
            return self.index > other.index
        return NotImplemented

    def __repr__(self):
        args = ", True" if self.ro else ""
        args += f", name={self.name!r}" if self.name else ""
        return f"state_property({self.index}{args})"

    def __str__(self):
        return self.name or super().__str__()


class param_property(property):
    """
    A property that simply mirrors a given param value.
    """

    is_param = True
    is_derived = False

    def __init__(self, name=None, ro=False, default=NOT_GIVEN):
        if callable(default):
            default = sk.extract_function(default)

        self.name = name
        self.default = default
        param = self

        def fget(self):
            try:
                return self._params[param.name].value
            except KeyError:
                if param.default is NOT_GIVEN:
                    name = type(self).__name__
                    raise AttributeError(f"{name} has no {param.name!r} attribute")
                value = param.default
                return value(self) if callable(value) else value

        def fset(self, value):
            self.set_param(param.name, value)

        args = (fget,) if ro else (fget, fset)
        super().__init__(*args)

    def __set_name__(self, owner, name):
        self.name = self.name or name


class param_transform(property):
    """
    A property that represents a transform into a parameter.
    """

    is_param = True
    is_derived = True

    def __init__(self, prop_name, read, write=None):
        read = sk.extract_function(read)

        if write is not None:
            write = sk.extract_function(write)

        def fget(self):
            try:
                value = self._params[prop_name].value
            except KeyError:
                value = self.get_param(prop_name)
            try:
                return read(value)
            except Exception as e:
                cls = type(e).__name__
                msg = f"error found when processing {prop_name!r}: {cls}{e}"
                raise ValueError(msg)

        def fset(self, value):
            value = write(value)
            self.set_param(prop_name, value)

        args = (fget, fset) if write else (fget,)
        super().__init__(*args)


def param_alias(prop, ro=False):
    """
    A simple alias to a parameter.
    """

    id_fn = lambda x: x
    return param_transform(prop, read=id_fn, write=None if ro else id_fn)


def inverse_transform(prop):
    """
    A param_transform for a parameter that is the inverse of the other.
    """

    return param_transform(prop, read=lambda x: 1 / x, write=lambda x: 1 / x)



# flake8: noqa
import datetime as dt
from datetime import date, datetime, time, timedelta

from sidekick import import_later as _imp
from sidekick import fn, X, Y, F, X_i, L, placeholder as _
import mundi
import mundi_demography as mdm
import mundi_healthcare as mhc

from .jupyter import *
from .packages import *
from .models import *
from .utils import *
from .diseases import *
from .formulas import *
from . import region

sk = _imp("sidekick")
h = _imp("hyperpython:h")
np = _imp("numpy")
pd = _imp("pandas")
plt = _imp("matplotlib.pyplot")
sns = _imp("seaborn")
sm = _imp("statsmodels.api")
smf = _imp("statsmodels.formula.api")
pydemic = _imp("pydemic")


#
# Useful constants in a Jupyter notebook
#
DAY = dt.timedelta(days=1)


def evil():
    """
    Enable sidekick's forbidden functional powers.
    """
    from sidekick.evil import forbidden_powers

    forbidden_powers()


def documentation_mode():
    """
    Monkey patch random things before executing a notebook that will be part of
    documentation.
    """
    from matplotlib.axes import SubplotBase

    SubplotBase.__repr__ = lambda _: ""



from pydemic.models import SIR
from pydemic.models.abstract_sir import AbstractSIR


class TestMetaClass:
    def test_meta_class_for_derived_classes(self):
        assert AbstractSIR.meta.model_name == "SIR"
        assert SIR.meta.model_name == "SIR"

    def test_meta_class_introspection(self):
        meta = SIR.meta

        # State variables
        assert meta.variables == ("susceptible", "infectious", "recovered")
        assert meta.ndim == 3
        assert meta.data_aliases == {
            **dict(zip("SIR", meta.variables)),
            "E": "infectious",
            "exposed": "infectious",
        }

        # Parameters
        assert meta.params.primary == {"R0", "infectious_period"}
        assert meta.params.alternative == {"gamma"}
        assert meta.params.all == {*meta.params.primary, *meta.params.alternative}



import sidekick as sk
from sidekick import placeholder as _

from pydemic.params import WrappedParams


class DiseaseParams(WrappedParams):
    """
    A wrapper for disease params.
    """

    gamma: float = sk.property(1 / _.infectious_period)
    sigma: float = sk.property(1 / _.incubation_period)

    Qs: float = sk.alias("prob_symptoms")
    Qsv: float = sk.alias("prob_severe")
    Qcr: float = sk.alias("prob_critical")
    CFR: float = sk.alias("case_fatality_ratio")
    IFR: float = sk.alias("infection_fatality_ratio")
    HFR: float = sk.alias("hospital_fatality_ratio")
    ICUFR: float = sk.alias("icu_fatality_ratio")



from typing import Union

import numpy as np
import statsmodels.api as sm

from ..types import Result


def smoothed_diff(
    data: np.ndarray,
    *,
    bias: float = 1e-4,
    damped: bool = True,
    linear: bool = False,
    fast_init: bool = False,
    retall: bool = False,
    **kwargs,
) -> Union[np.ndarray, Result]:
    """
    Load a series of values and use Holt's algorithm to smooth the finite
    differences.

    Args:
        values:
            Input data.
        damped:
            If True (default), enable damping of the exponential trend.
        linear:
            If True, make the trend linear instead of exponential.
        bias:
            Holt's algorithm require strictly positive values (i.e., greater than,
            but not equal to zero). This parameter controls a bias term that is
            added to the baseline to prevent the existence of null values.
            This transformation is not used if linear=True.
        fast_init:
            If True, prevents using the slower brute force search during
            initialization of the model.
        retall:
            If True, return a named tuple (value, info) with both the computed
            smoothed array and the statistical model used to calibrate the
            parameters.
        **kwargs:

    Returns:
        Either a :cls:`np.ndarray` of smoothed differences or a :cls:`result` value.
    """
    values = np.asarray(data)
    maximum_value = values.max()
    values = np.diff(values, prepend=0.0)

    # Guarantee that there are no zero values in the
    # dataset
    if not linear:
        values = np.maximum(values, 0.0)
        S = values.sum()
        values *= 1 - bias
        values += bias * S

    # Choose empirically validated values
    kwargs.setdefault("smoothing_level", 1 / 5)

    # Holt exponential smoothing
    holt = sm.tsa.Holt(values, exponential=not linear, damped=damped)
    res = holt.fit(use_brute=not fast_init, **kwargs)
    out = res.fittedvalues
    out *= maximum_value / out.sum()
    return Result(out, res) if retall else out



import pandas as pd

from .base import RegionProperty
from .. import fitting as fit
from ..diseases import disease as get_disease, DiseaseParams
from ..properties.decorators import cached
from ..types import ValueStd


class PydemicProperty(RegionProperty):
    """
    Implements the "pydemic" property mokey-patched into mundi region objects.
    """

    __slots__ = ()

    @cached(ttl="epidemic_curve")
    def epidemic_curve(self, disease=None, **kwargs) -> pd.DataFrame:
        """
        Return epidemic curves for region.

        See Also:
            :method:`pydemic.diseases.Disease.epidemic_curve`
        """
        disease = get_disease(disease)
        return disease.epidemic_curve(self.region, **kwargs)

    def disease_params(self, disease=None, **kwargs) -> DiseaseParams:
        """
        Return an object with all disease params associated with region.
        """
        disease = get_disease(disease)
        return disease.params(region=self.region, **kwargs)

    @cached(ttl="epidemic_curve")
    def estimate_Rt(self, model, disease=None, **kwargs) -> pd.DataFrame:
        """
        Compute R(t) from the epidemic curves. This is function is just a
        shortcut to :func:`fit.estimate_Kt` that automatically binds to the
        current region.
        """
        return self._estimate_R(fit.estimate_Rt, model, disease, **kwargs)

    @cached(ttl="epidemic_curve")
    def estimate_R0(self, model, disease=None, **kwargs) -> ValueStd:
        """
        Compute R0 from the epidemic curves. This is function is just a
        shortcut to :func:`fit.estimate_Kt` that automatically binds to the
        current region.
        """
        return self._estimate_R(fit.estimate_R0, model, disease, **kwargs)

    def _estimate_R(self, func, model, disease, **kwargs):
        disease = get_disease(disease)
        curves = self.epidemic_curve(disease)
        if isinstance(model, str) and "params" not in kwargs:
            kwargs["params"] = disease.params(region=self.region)
        return func(model, curves, **kwargs)

    @cached(ttl="epidemic_curve")
    def estimate_Kt(self, disease=None, **kwargs) -> pd.DataFrame:
        """
        Compute K(t) from the epidemic curves. This is function is just a
        shortcut to :func:`fit.estimate_Kt` that automatically binds to the
        current region.
        """
        return self._estimate_K(fit.estimate_Kt, disease, **kwargs)

    @cached(ttl="epidemic_curve")
    def estimate_K(self, disease=None, **kwargs) -> ValueStd:
        """
        Compute K from the epidemic curves. This is function is just a
        shortcut to :func:`fit.estimate_K` that automatically binds to the
        current region.
        """
        return self._estimate_K(fit.estimate_K, disease, **kwargs)

    def _estimate_K(self, func, disease, **kwargs):
        curves = self.epidemic_curve(disease)
        return func(curves, **kwargs)



from abc import ABCMeta

from ..mixins import Meta


class ModelMeta(ABCMeta):
    """
    Metaclass for model classes.
    """

    meta: "Meta"

    def __init__(cls, name, bases, ns):
        meta = ns.pop("Meta", None)
        super().__init__(name, bases, ns)

        cls.meta = Meta.from_arguments(cls, meta)



# flake8: noqa
from .model import Model
from .esir import eSIR
from .eseir import eSEIR
from .eseair import eSEAIR
from .sir import SIR
from .seir import SEIR
from .seair import SEAIR
from .ode_model import ODEModel



from abc import ABC

import numpy as np

from .model import Model


class ODEModel(Model, ABC):
    """
    Base class for all models that uses ordinary differential equations.
    """

    integration_method = "RK4"
    sub_steps = 4

    def diff(self, x: np.ndarray, t: float) -> np.ndarray:
        """
        Derivative function for the ODE.

        Must be implemented in subclasses.
        """
        raise NotImplementedError

    def integration_step(self, x, t, dt, method=None):
        """
        A single RK4 iteration step.
        """
        method = method or self.integration_method
        if method == "RK4":
            k1 = self.diff(x, t)
            k2 = self.diff(x + 0.5 * dt * k1, t + 0.5 * dt)
            k3 = self.diff(x + 0.5 * dt * k2, t + 0.5 * dt)
            k4 = self.diff(x + 1.0 * dt * k3, t + 1.0 * dt)
            return x + (k1 + 2 * k2 + 2 * k3 + k4) * (dt / 6)
        else:
            raise ValueError(f"unknown integration method: {method!r}")

    def run_to_fill(self, data, times):
        x = self.state
        for i, t in enumerate(times):
            dt = (t - self.time) / self.sub_steps
            for j in range(self.sub_steps):
                x = self.integration_step(x, t, dt)
            self.time = t
            self.state = x
            data[i] = x
        return x



# flake8: noqa
import sidekick as _sk

from .datetime import today, now
from .formatting import (
    fmt,
    pc,
    p10k,
    p100k,
    pm,
    indent,
    slugify,
    safe_int,
    format_args,
    file_type_display_name,
)
from .functions import interpolant, lru_safe_cache, coalesce, maybe_run
from .json import to_json
from .properties import (
    state_property,
    param_property,
    param_transform,
    param_alias,
    inverse_transform,
)
from .sequences import (
    rpartition,
    flatten_dict,
    unflatten_dict,
    extract_keys,
    sliced,
    as_seq,
    is_seq,
)
from .dataframe import trim_zeros, force_monotonic
from .timeseries import accumulate_weekly, day_of_week, weekday_name, trim_weeks

not_implemented = lambda *args: _sk.error(NotImplementedError)



from .base_attr import BaseAttr


class Results(BaseAttr):
    """
    Results objects store dynamic information about a simulation.

    Results data is not necessarily static throughout the simulation. It can
    include values like total death toll, number of infected people, etc. It also
    distinguishes from information stored in the model mapping interface in
    that it does not include time series.

    Most information available as ``m[<param>]`` will also be available as
    ``m.results[<param>]``. While the first typically include the whole time
    series for the object, the second typically correspond to the last value
    in the time series.
    """

    __slots__ = ()
    _method_namespace = "results"

    @property
    def _cache(self):
        model = self.model
        if model.iter != model._results_dirty_check:
            model._results_cache.clear()
            model._results_dirty_check = model.iter
        return model._results_cache



from .info import Info
from .meta_info import Meta
from .params_info import ParamsInfo
from .results import Results
from .with_data import WithDataMixin, WithDataModelMixin
from .with_info import WithInfoMixin
from .with_params import WithParamsMixin, WithDynamicParamsMixin
from .with_region import WithRegionDemography
from .with_results import WithResultsMixin



import datetime
from copy import copy
from types import MappingProxyType
from typing import Sequence, Callable, Mapping, TYPE_CHECKING, Union

import numpy as np
import pandas as pd
import sidekick as sk

from .clinical_acessor import Clinical
from .metaclass import ModelMeta
from .. import fitting as fit
from .. import formulas
from ..diseases import Disease, DiseaseParams, disease as get_disease
from ..logging import log
from ..mixins import (
    Meta,
    WithParamsMixin,
    WithDataModelMixin,
    WithInfoMixin,
    WithResultsMixin,
    WithRegionDemography,
)
from ..packages import plt
from ..utils import today, not_implemented, extract_keys, param_property

NOW = datetime.datetime.now()
TODAY = datetime.date(NOW.year, NOW.month, NOW.day)
DAY = datetime.timedelta(days=1)
pplt = sk.import_later("..plot", package=__package__)

if TYPE_CHECKING:
    from ..model_group import ModelGroup
    from pydemic_ui.model import UIProperty


class Model(
    WithDataModelMixin,
    WithInfoMixin,
    WithResultsMixin,
    WithParamsMixin,
    WithRegionDemography,
    metaclass=ModelMeta,
):
    """
    Base class for all models.
    """

    meta: Meta

    class Meta:
        model_name = "Model"
        data_aliases = {}

    # Initial values
    state: np.ndarray = None
    initial_cases: float = sk.lazy(lambda self: self._initial_cases())
    initial_infected: float = sk.lazy(lambda self: self._initial_infected())

    # Initial time
    date: datetime.date = None
    time: float = 0.0
    iter: int = sk.property(lambda m: len(m.data))
    dates: pd.DatetimeIndex = sk.property(lambda m: m.to_dates(m.times))
    times: pd.Index = sk.property(lambda m: m.data.index)

    # Common epidemiological parameters
    R0: float = param_property("R0", default=2.0)
    K = sk.property(not_implemented)
    duplication_time = property(lambda self: np.log(2) / self.K)

    # Special accessors
    clinical: Clinical = property(lambda self: Clinical(self))
    clinical_model: type = None
    clinical_params: Mapping = MappingProxyType({})
    disease: Disease = None
    disease_params: DiseaseParams

    @property
    def ui(self) -> "UIProperty":
        try:
            from pydemic_ui.model import UIProperty
        except ImportError as ex:
            log.warn(f"Could not import pydemic_ui.model: {ex}")
            msg = (
                "must have pydemic-ui installed to access the model.ui attribute.\n"
                "Please 'pip install pydemic-ui' before proceeding'"
            )
            raise RuntimeError(msg)
        return UIProperty(self)

    @classmethod
    def create(cls, params=None):
        new = object.__new__(cls)
        new.set_params(params)
        return new

    def __init__(
        self, params=None, *, run=None, name=None, date=None, clinical=None, disease=None, **kwargs
    ):
        self.name = name or f"{type(self).__name__} model"
        self.date = pd.to_datetime(date or today())
        self.disease = get_disease(disease)
        self._initialized = False

        # Fix demography
        demography_opts = WithRegionDemography._init_from_dict(self, kwargs)
        self.disease_params = self.disease.params(**demography_opts)

        # Init other mixins
        WithParamsMixin.__init__(self, params, keywords=kwargs)
        WithInfoMixin.__init__(self)
        WithResultsMixin.__init__(self)
        WithDataModelMixin.__init__(self)

        if clinical:
            clinical = dict(clinical)
            self.clinical_model = clinical.pop("model", None)
            self.clinical_params = clinical

        for k, v in kwargs.items():
            if hasattr(self, k):
                try:
                    setattr(self, k, v)
                except AttributeError:
                    name = type(self).__name__
                    msg = f"cannot set '{k}' attribute in '{name}' model"
                    raise AttributeError(msg)
            else:
                raise TypeError(f"invalid arguments: {k}")

        if run is not None:
            self.run(run)

    def __str__(self):
        return self.name

    def _initial_cases(self):
        raise NotImplementedError("must be implemented in subclass")

    def _initial_infected(self):
        raise NotImplementedError("must be implemented in subclass")

    #
    # Pickling and copying
    #
    # noinspection PyUnresolvedReferences
    def copy(self, **kwargs):
        """
        Copy instance possibly setting new values for attributes.

        Keyword Args:
            All keyword arguments are used to reset attributes in the copy.

        Examples:
            >>> m.copy(R0=1.0, name="Stable")
            <SIR(name="Stable")>
        """

        cls = type(self)
        data = self.__dict__.copy()
        params = data.pop("_params")
        data.pop("_results_cache")

        new = object.__new__(cls)
        for k in list(kwargs):
            if k in data:
                data[k] = kwargs.pop(k)

        new._params = copy(params)
        new._results_cache = {}
        new.__dict__.update(copy(data))

        for k, v in kwargs.items():
            setattr(new, k, v)

        return new

    def split(self, n=None, **kwargs) -> "ModelGroup":
        """
        Create n copies of model, each one may override a different set of
        parameters and return a ModelGroup.

        Args:
            n:
                Number of copies in the resulting list. It can also be a sequence
                of dictionaries with arguments to pass to the .copy() constructor.

        Keyword Args:
            Keyword arguments are passed to the `.copy()` method of the model. If
            the keyword is a sequence, it applies the n-th component of the sequence
            to the corresponding n-th model.
        """

        from ..model_group import ModelGroup

        if n is None:
            for k, v in kwargs.items():
                if not isinstance(v, str) and isinstance(v, Sequence):
                    n = len(v)
                    break
            else:
                raise TypeError("cannot determine the group size from arguments")

        if isinstance(n, int):
            options = [{} for _ in range(n)]
        else:
            options = [dict(d) for d in n]
        n: int = len(options)

        # Merge option dicts
        for k, v in kwargs.items():
            if not isinstance(v, str) and isinstance(v, Sequence):
                xs = v
                m = len(xs)
                if m != n:
                    raise ValueError(
                        f"sizes do not match: "
                        f"{k} should be a sequence of {n} "
                        f"items, got {m}"
                    )
                for opt, x in zip(options, xs):
                    opt.setdefault(k, x)
            else:
                for opt in options:
                    opt.setdefault(k, v)

        # Fix name
        for opt in options:
            try:
                name = opt["name"]
            except KeyError:
                pass
            else:
                opt["name"] = name.format(n=n, **opt)

        return ModelGroup(self.copy(**opt) for opt in options)

    def split_children(self, options=MappingProxyType({}), **kwargs) -> "ModelGroup":
        """
        Similar to split, but split into the children of the given class.

        Args:
            options:
                A mapping between region or region id
        """

        from ..model_group import ModelGroup

        if self.region is None:
            raise ValueError("model is not bound to a region")

        for k in self._params:
            if k not in kwargs:
                kwargs[k] = self.get_param(k)

        for attr in ("disease",):
            kwargs.setdefault(attr, getattr(self, attr))

        return ModelGroup.from_children(self.region, type(self), options, **kwargs)

    def reset(self, date: Union[datetime.date, float] = None, **kwargs):
        """
        Return a copy of the model setting the state to the final state. If a
        positional "date" argument is given, reset to the state to the one in the
        specified date.

        Args:
            date (float or date):
                An optional float or datetime selecting the desired date.

        Keyword Args:
            Additional keyword arguments are handled the same way as the
            :method:`copy` method.
        """
        if date is None:
            date = self.date
            time = self.time
        elif isinstance(date, (float, int)):
            time = float(date)
            date = self.to_date(date)
        else:
            time: float = self.to_time(date)

        kwargs["data"] = self.data.loc[[time]]
        kwargs["date"] = date
        kwargs["state"] = kwargs["data"].iloc[0].values
        kwargs["time"] = 1
        return self.copy(**kwargs)

    def trim_dates(self, start=0, end=None):
        """
        Trim data in model to the given interval specified by start and end
        dates or times.

        Args:
            start (int or date):
                Starting date. If not given, start at zero.
            end (int or date):
                End date. If not given, select up to the final date.
        """
        start = int(start or 0)
        end = int(end or self.time)
        new = self.copy(
            date=self.to_date(start),
            data=self.data.iloc[start:end].reset_index(drop=True),
            time=end - start,
            state=self.data.iloc[end].values,
        )
        return new

    def epidemic_model_name(self):
        """
        Return the epidemic model name.
        """
        return self.meta.model_name

    #
    # Initial conditions
    #
    def set_ic(self, state=None, **kwargs):
        """
        Set initial conditions.
        """
        if self.state is None:
            if state is None:
                state = self.initial_state(**kwargs)
            self.state = np.array(state, dtype=float)

        alias = self.meta.data_aliases
        for k, v in list(kwargs.items()):
            if k in alias:
                del kwargs[k]
                kwargs[alias[k]] = v
        components = extract_keys(self.meta.variables, kwargs)

        for k, v in components.items():
            idx = self.meta.get_variable_index(k)
            self.state[idx] = v

    def set_data(self, data):
        """
        Force a dataframe into simulation state.
        """
        data = data.copy()
        data.columns = [self.meta.data_aliases.get(c, c) for c in data.columns]

        self.set_ic(state=data.iloc[0])
        self.data = data.reset_index(drop=True)
        self.time = len(data) - 1
        self.date = data.index[-1]
        self.state[:] = data.iloc[-1]
        self.info["observed.dates"] = data.index[[0, -1]]
        self._initialized = True

    def set_cases(self, curves=None, adjust_R0=False, save_cases=False):
        """
        Initialize model from a dataframe with the deaths and cases curve.

        This curve is usually the output of disease.epidemic_curve(region), and is
        automatically retrieved if not passed explicitly and the region of the model
        is set.

        Args:
            curves:
                Dataframe with cumulative ["cases", "deaths"] columns. If not given,
                or None, fetches from disease.epidemic_curves(info)
            adjust_R0:
                If true, adjust R0 from the observed cases.
            save_cases:
                If true, save the cases curves into the model.info["observed.cases"] key.
        """

        if curves is None:
            if self.region is None or self.disease is None:
                msg = 'must provide both "region" and "disease" or an explicit cases ' "curve."
                raise ValueError(msg)
            curves = self.region.pydemic.epidemic_curve(self.disease, real=True)

        if adjust_R0:
            method = "RollingOLS" if adjust_R0 is True else adjust_R0
            Re, _ = value = fit.estimate_R0(self, curves, Re=True, method=method)
            assert np.isfinite(Re), f"invalid value for R0: {value}"
            self.R0 = Re

        # Save notification it in the info dictionary for reference
        if "cases_observed" in curves:
            tf = curves.index[-1]
            rate = curves.loc[tf, "cases_observed"] / curves.loc[tf, "cases"]
        else:
            rate = 1.0
        self.info["observed.notification_rate"] = rate

        # Save simulation state from data
        model = self.epidemic_model_name()
        curve = fit.cases(curves)
        data = fit.epidemic_curve(model, curve, self)
        self.set_data(data)
        try:
            pred_cases = self["cases"]
        except KeyError:
            log.warn("Could not determine model cases")
        else:
            self.initial_cases -= pred_cases.iloc[-1] - curve.iloc[-1] + 1000

        if adjust_R0:
            self.R0 /= self["susceptible:final"] / self.population
            self.info["observed.R0"] = self.R0

        # Optionally save cases curves into the info dictionary
        if save_cases:
            key = "observed.cases" if save_cases is True else save_cases
            df = curves.copy()
            df["cases"] = curve
            df["cases_raw"] = curves["cases"]
            self.info[key] = df

    def initial_state(self, cases=None, **kwargs):
        """
        Create the default initial vector for model.
        """
        if cases is not None:
            kwargs.setdefault("population", self.population)
            return formulas.initial_state(self.epidemic_model_name(), cases, self, **kwargs)
        return self._initial_state()

    def _initial_state(self):
        raise NotImplementedError

    def initialize(self):
        """
        Force initialization.
        """
        if not self._initialized:
            self.set_ic()
            self.data = make_dataframe(self)
        self._initialized = True

    #
    # Running simulation
    #
    def run(self, time):
        """
        Runs the model for the given duration.
        """
        steps = int(time)
        self.initialize()

        if time == 0:
            return

        _, *shape = self.data.shape

        ts = self.time + 1.0 + np.arange(steps)

        data = np.zeros((steps, *shape))
        date = self.date

        if self.info.get("event.simulation_start") is None:
            self.info.save_event("simulation_start")

        self.run_to_fill(data, ts)
        extra = pd.DataFrame(data, columns=self.data.columns, index=ts)

        self.data = pd.concat([self.data, extra])
        self.date = date + time * DAY
        self.time = ts[-1]
        self.state = data[-1]

    def run_to_fill(self, data, times):
        """
        Run simulation to fill pre-allocated array of data.
        """
        raise NotImplementedError

    def run_until(self, condition: Callable[["Model"], bool]):
        """
        Run until stop condition is satisfied.

        Args:
            condition:
                A function that receives a model and return True if stop
                criteria is satisfied.
        """
        raise NotImplementedError

    #
    # Utility methods
    #
    def to_dates(self, times: Sequence[int], start_date=None) -> pd.DatetimeIndex:
        """
        Convert an array of numerical times to dates.

        Args:
            times:
                Sequence of times.
            start_date:
                Starting date. If not given, uses the starting date for
                simulation.
        """
        dates: pd.DatetimeIndex

        if isinstance(times, pd.DatetimeIndex):
            return times
        if start_date is None:
            start_date = self.date - self.time * DAY

        # noinspection PyTypeChecker
        return pd.to_datetime(times, unit="D", origin=start_date)

    def to_date(self, time: Union[float, int]) -> datetime.date:
        """
        Convert a single instant to the corresponding datetime
        """
        return pd.to_datetime(time - self.time, unit="D", origin=self.date)

    def to_times(self, dates: Sequence, start_date=None) -> np.ndarray:
        """
        Convert an array of numerical times to dates.

        Args:
            dates:
                Sequence of dates.
            start_date:
                Starting date. If not given, uses the starting date for
                simulation.
        """
        if start_date is None:
            start_date = self.date - self.time * DAY
        data = [(date - start_date).days for date in dates]
        return np.array(data) if data else np.array([], dtype=int)

    def to_time(self, date, start_date=None) -> float:
        """
        Convert date to time.
        """
        if start_date is None:
            return self.to_time(date, self.date) - self.time
        return float((date - start_date).days)

    def get_times(self, idx=None):
        """
        Get times possibly sliced by an index.
        """
        if idx is None:
            return self.times
        else:
            return self.times[idx]

    def get_data_cases(self, idx):
        raise NotImplementedError

    #
    # Plotting and showing information
    #
    def plot(
        self,
        components=None,
        *,
        ax=None,
        logy=False,
        show=False,
        dates=False,
        legend=True,
        grid=True,
    ):
        """
        Plot the result of simulation.
        """
        ax = ax or plt.gca()
        kwargs = {"logy": logy, "ax": ax, "grid": grid, "legend": legend}

        def get_column(col):
            if dates:
                col += ":dates"
            data = self[col]
            return data

        components = self.meta.variables if components is None else components
        for col in components:
            data = get_column(col)
            data.plot(**kwargs)
        if show:
            plt.show()


def make_dataframe(model: Model):
    """
    Create the initial dataframe for the given model.
    """
    data = [model.state]
    cols = model.meta.variables
    index = [model.time]
    return pd.DataFrame(data, columns=cols, index=index)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pydemic',
 'pydemic.api',
 'pydemic.cli',
 'pydemic.clinical_models',
 'pydemic.diseases',
 'pydemic.fitting',
 'pydemic.formulas',
 'pydemic.mixins',
 'pydemic.model_group',
 'pydemic.models',
 'pydemic.params',
 'pydemic.plot',
 'pydemic.properties',
 'pydemic.region',
 'pydemic.reports',
 'pydemic.utils']

package_data = \
{'': ['*'],
 'pydemic': ['databases/*',
             'databases/diseases/*',
             'databases/diseases/covid-19/*',
             'databases/diseases/empty/*'],
 'pydemic.models': ['.idea/*', '.idea/inspectionProfiles/*']}

install_requires = \
['mundi',
 'mundi-demography',
 'mundi-healthcare',
 'joblib',
 'click~=7.0',
 'matplotlib~=3.0',
 'numpy~=1.0',
 'pandas~=1.0',
 'requests~=2.0',
 'seaborn~=0.10',
 'statsmodels']

extras_require = \
{'dev': ['black', 'invoke~=1.4', 'flake8', 'pytest~=5.4', 'pytest-cov~=2.8']}

setup(name='pydemic-models',
      version='0.1.10',
      description='Pydemic',
      author='Fábio Macêdo Mendes',
      author_email='fabiomacedomendes@gmail.com',
      url='http://github.com/pydemic/pydemic',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



from pandas.testing import assert_series_equal

from mundi import region
from pydemic.region.multi_region import CompositeRegion


class TestCompositeRegion:
    def test_composite_region_with_single_element(self):
        bsb = region("BR-5300108")
        reg = CompositeRegion([bsb])
        for attr in ("population", "name", "type", "subtype"):
            assert getattr(bsb, attr) == getattr(reg, attr)

    def test_arbitrary_composite_region(self):
        bsb = region("BR-5300108")
        sp = region("BR-3550308")
        reg = CompositeRegion([bsb, sp])

        assert reg.population == bsb.population + sp.population
        assert_series_equal(reg.age_distribution, bsb.age_distribution + sp.age_distribution)



import operator as op
from typing import Sequence

import numpy as np
import pandas as pd

_ = lambda x: x
WEEKDAY_NAMES = [
    _("Monday"),
    _("Tuesday"),
    _("Wednesday"),
    _("Thursday"),
    _("Friday"),
    _("Saturday"),
    _("Sunday"),
]


def trim_weeks(data: pd.DataFrame, week_start=0) -> pd.DataFrame:
    """
    Given a dataframe indexed with datetime values, trim both ends to correspond
    to whole weeks.

    Args:
        data:
            A dataframe or series object indexed with dates.
        week_start:
            First day of each week: Monday = 0, Tuesday = 1, and so on.
    """

    weekday = day_of_week(data)
    first_day = weekday.iloc[0]
    if first_day != week_start:
        idx = (week_start - first_day) % 7
        data = data.iloc[idx:]

    week_end = (week_start + 6) % 7
    last_day = weekday.iloc[-1]
    if last_day != week_end:
        idx = len(data) - (last_day - week_end) % 7
        data = data.iloc[:idx]

    return data


def accumulate_weekly(data: pd.DataFrame, method="strict", week_start=0, acc=lambda x: x.sum()):
    """
    Accumulate values by week.

    Args:
        data:
            A dataframe with pandas datetimes in the index.
        method:
            Governs how the function handles partial weeks.
            * 'strict' (default): Raise an error.
            * 'trim-start': Remove additional days from the beginning of the time series.
            * 'trim-end': Remove additional days from the end of the time series.
            * 'trim': Trim both sides to make all weeks start in a predictable day.
        week_start:
            If method is 'trim', controls which day is considered to be the first day
            of the week. Monday = 0, Tuesday = 1, and so on.
        acc:
            Accumulator function.
    """
    if method == "trim":
        data = trim_weeks(data, week_start)
    elif method == "strict":
        if len(data) % 7 != 0:
            raise ValueError("series contains partial weeks")
    elif method == "trim-start":
        extra = len(data) % 7
        if extra:
            data = data.iloc[:-extra]
    elif method == "trim-end":
        extra = len(data) % 7
        if extra:
            data = data.iloc[extra:]
    else:
        raise ValueError(f"invalid method: {method!r}")

    by_week = []
    n_weeks = len(data) // 7
    for week in range(n_weeks):
        total = acc(data.iloc[7 * week : 7 * (week + 1)])
        by_week.append(total)
    return pd.DataFrame(by_week)


def day_of_week(df: pd.DataFrame, col=None) -> pd.Series:
    """
    Return a series for weekdays from input.

    Args:
        df:
            Dataframe or series object.
        col:
            Name of the column with datetime values. If not given, assumes to
            be the index.
    """
    if col is None:
        data = df.index
    else:
        data = df[col]
    data = [*map(op.attrgetter("dayofweek"), data)]
    return pd.Series(data, index=df.index)


def weekday_name(idx, translate=str):
    """
    Return name for weekday from index.

    Input can be an scalar or a sequence.

    Notes:
          Monday = 0, Tuesday = 1, and so on.
    """
    _ = translate or str

    if isinstance(idx, Sequence):
        names = WEEKDAY_NAMES
        print(idx)
        return np.array([_(names[i]) for i in idx])
    return _(WEEKDAY_NAMES[idx])



from abc import ABC
from collections import defaultdict
from typing import TYPE_CHECKING

from .info import Info

if TYPE_CHECKING:
    from ..models import Model  # noqa: F401


class WithInfoMixin(ABC):
    """
    Subclasses that adopt this mixin gain a "info" attribute that expose
    useful information about its owner.

    In the context of simulation models, it shows useful static information
    about the simulation.
    """

    INFO_DEMOGRAPHY_KEYS = ("population", "age_distribution", "age_pyramid")
    INFO_REGION_KEYS = (
        "population",
        "age_distribution",
        "age_pyramid",
        "icu_capacity",
        "hospital_capacity",
    )

    def __init__(self):
        self._info_cache = defaultdict(dict)

    @property
    def info(self) -> Info:
        return Info(self)

    def get_info_keys_demography(self):
        """
        Yield keys for the info["demography"] dict.
        """
        yield from self.INFO_DEMOGRAPHY_KEYS

    def get_info_value_demography(self: "Model", key):
        """
        Return value for model.info["demography.<key>"] queries.
        """
        if key == "population":
            return self.data.iloc[0].sum()
        elif key == "age_distribution":
            return self.age_distribution
        elif key == "age_pyramid":
            return self.age_pyramid
        elif key.startswith("population("):
            _, _, cmd = key.partition("(")
            cmd = cmd.rstrip(")")
            return get_population_fraction(cmd, self.age_distribution)
        else:
            raise ValueError(f"unknown argument: {key!r}")

    def get_info_keys_region(self: "Model"):
        """
        Yield keys for the result["region"] dict.
        """
        if self.region is None:
            return
        yield from self.INFO_REGION_KEYS

    def get_info_value_region(self: "Model", key):
        """
        Handle model.result["region.*"] queries.
        """
        if key in self.INFO_REGION_KEYS:
            return getattr(self.region, key)
        return KeyError(key)

    def get_info_keys_disease(self: "Model"):
        """
        Yield keys for the result["disease"] dict.
        """
        yield from self.disease_params

    def get_info_value_disease(self: "Model", key):
        """
        Return value for model.result["disease.<key>"] queries.
        """
        return getattr(self.disease_params, key)


def get_population_fraction(cmd, distribution):
    """
    Given a command like 60+, <10, 50-59, etc, return the fraction of population
    in the given interval.
    """

    total = distribution.sum()
    if cmd.endwith("+"):
        pos = int(cmd[:-1])
        return distribution.loc[pos:].sum() / total
    elif cmd.startswith("<"):
        pos = int(cmd[1:])
        return distribution.loc[:pos].sum() / total
    else:
        a, b = map(int, cmd.split("-"))
        b += 1
        return distribution.loc[a:b].sum() / total



import re
from functools import singledispatch, lru_cache
from gettext import gettext as _
from itertools import chain
from math import log10
from numbers import Number
from types import MappingProxyType

import numpy as np

from .sequences import rpartition

N_RE = re.compile(r"(-?)(\d+)(\.\d{,2})?\d*")
identity = lambda x: x
HUMANIZED_SUFFIXES = (
    (0.1, 1, identity, "{:.04n}", ""),
    (1e2, 1, identity, "{:.03n}", ""),
    (1e3, 1, identity, "{:.2n}", ""),
    (1e6, 1, int, "{:n}", ""),
    (1e8, 1e6, identity, "{:.3n}", _("M")),
    (1e9, 1e6, identity, "{:.2n}", _("M")),
    (1e11, 1e9, identity, "{:.3n}", _("B")),
    (1e12, 1e9, identity, "{:.2n}", _("B")),
    (1e13, 1e12, identity, "{:.3n}", _("T")),
    (1e14, 1e12, identity, "{:.2n}", _("T")),
)


@singledispatch
def fmt(x, empty="-", role=None):
    """
    Format element to a humanized form.
    """
    if x is None:
        return empty
    return str(x)


@fmt.register(Number)
def __(x, empty="-", role=None):
    if role is None:
        return format_number(x, empty)
    elif role in ("pc", "%"):
        return pc(x, empty)
    elif role == "pm":
        return pm(x, empty)
    elif role == "p10k":
        return p10k(x, empty)
    elif role == "p100k":
        return p100k(x, empty)
    else:
        raise ValueError(f"invalid role: {role!r}")


def format_number(n, empty="-"):
    """
    Heuristically choose best format option for number.
    """
    if n is None:
        return empty

    if n == float("inf"):
        return _("infinity")
    try:
        return ", ".join(map(fmt, n))
    except TypeError:
        pass
    m = abs(n)

    if int(n) == n and m < 1e6:
        return "{:n}".format(int(n))

    if m < 0.01:
        e = int(-log10(m)) + 1
        return "{:.3n}".format(n * 10 ** e) + "e-%02d" % e
    for k, div, fn, fmt_, suffix in HUMANIZED_SUFFIXES:
        if m < k:
            dec = fn(m) / div
            dec -= int(dec)
            dec = fmt_.format(1 + dec)[1:]
            prefix = "{:n}".format(int(fn(n) / div))
            return prefix + dec + suffix
    return "{:.2n}".format(n)


def _fmt_aux(n, suffix=""):
    m = N_RE.match(str(n))
    sign, number, decimal = m.groups()
    return sign + _fix_int(number, 3) + (decimal or "") + suffix


def _fix_int(s, n):
    return ",".join(map("".join, rpartition(s, n)))


def pc(n, empty="-"):
    """
    Write number as percentages.
    """
    if n is None or np.isnan(n):
        return empty
    elif n == 0:
        return "0.0%"
    elif np.isinf(n):
        return _("inf") if n > 0 else _("-inf")
    return format_number(100 * n) + "%"


def pm(n, empty="-"):
    """
    Write number as parts per thousand.
    """
    if n is None or np.isnan(n):
        return empty
    elif n == 0:
        return "0.0‰"
    elif np.isinf(n):
        return _("inf") if n > 0 else _("-inf")
    return format_number(1000 * n) + "‰"


def p10k(n, empty="-"):
    """
    Write number as parts per ten thousand.
    """
    if n is None or np.isnan(n):
        return empty
    elif n == 0:
        return "0.0‱"
    elif np.isinf(n):
        return _("inf") if n > 0 else _("-inf")
    return format_number(10000 * n) + "‱"


def p100k(n, empty="-"):
    """
    Write number as parts per 100 thousand.
    """
    if n is None or np.isnan(n):
        return empty
    elif n == 0:
        return "0.0"
    elif np.isinf(n):
        return _("inf") if n > 0 else _("-inf")
    return format_number(100000 * n) + "/100k"


def indent(st, indent=4):
    """
    Indent string.
    """
    if isinstance(indent, int):
        indent = " " * indent
    return "".join(indent + ln for ln in st.splitlines(keepends=True))


def slugify(st, suffixes=(), prefixes=()):
    """
    Simplify string into a slug.

    If a list of slugified suffixes or prefixes is given, they are removed from
    the resulting string.
    """
    st = st.lower()
    st = re.sub(r"[\s_]+", "-", st)
    st = re.sub(r"[^\w-]", "", st)
    if suffixes:
        regex = "|".join(map(re.escape, suffixes))
        re.sub(fr"{regex}$", "", st)
    if prefixes:
        regex = "|".join(map(re.escape, prefixes))
        re.sub(fr"^{regex}", "", st)
    return st


def safe_int(x, default=0):
    """
    Convert float to int, with a fallback value or NaNs and Infs.
    """

    try:
        return int(x)
    except ValueError:
        if not np.isfinite(x):
            return default
        raise


def format_args(*args, **kwargs):
    """
    Return a nice string representation of the given arguments.

    Examples:
        >>> format_args(1, 2, op="sum")
        "1, 2, op='sum'"
    """
    repr_args = map(repr, args)
    repr_kwargs = (f"{k}={v!r}" for k, v in kwargs.items())
    return ", ".join(chain(repr_args, repr_kwargs))


#
# Display name functions
#
def file_type_display_name(ext, mapping=None):
    """
    Humanized representation of a file name extension.
    """
    if mapping is None:
        mapping = _file_type_display_name_mapping()

    return str(mapping.get(ext, ext.upper()))


@lru_cache(1)
def _file_type_display_name_mapping():
    return MappingProxyType(
        {
            "csv": _("CSV"),
            "csv.gz": _("zipped CSV"),
            "pkl": _("Python pickle"),
            "pkl.gz": _("Python pickle"),
            "xls": _("Excel sheet"),
            "xlsx": _("Excel sheet"),
        }
    )



import click
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np


@click.command()
@click.argument("code")
@click.option("--api", "-a", default="auto", help="Force some specific API")
@click.option("--output", "-o", help="Save results to the given output")
@click.option("--plot", "-p", is_flag=True, help="Plot results")
@click.option("--log", "-l", is_flag=True, help="Use logarithmic scale")
@click.option("--daily", "-d", is_flag=True, help="Show new cases per day")
@click.option("--grid", "-g", is_flag=True, help="Add grid to the graph")
def covid19_api_downloader(code, api, output, plot, log, daily, grid):
    """
    Load data of COVID-19 cases and deaths.
    """
    from ..api.covid19 import epidemic_curve

    df = epidemic_curve(code, api)
    df.index = pd.to_datetime(df.index)
    df = df.sort_index()

    # Daily cases?
    if daily:
        data = np.diff(df.values, prepend=0, axis=0)
        df = pd.DataFrame(data, columns=df.columns, index=df.index)

    # Plotting
    if plot:
        df.plot.bar(logy=log, width=0.8) if daily else df.plot(logy=log)
        if grid:
            plt.grid()
        plt.title(code)
        plt.show()

    # Save output file
    def exts(lst):
        for ext in lst:
            if output.endswith("." + ext):
                return True
        return False

    if not output:
        if not plot:
            print(df)
    elif exts(["csv", "csv.gz"]):
        df.to_csv(output)
    elif exts(["csv", "csv.gz"]):
        df.to_pickle(output)
    elif exts(["xls", "xlsx"]):
        df.to_excel(output)
    else:
        raise SystemExit(f"Invalid output file: {output!r}")



import datetime


def today(n=0) -> datetime.date:
    """
    Return the date today.
    """
    today = now().date()
    if n:
        return today + datetime.timedelta(days=n)
    return today


def now() -> datetime.datetime:
    """
    Return a datetime timestamp.
    """
    return datetime.datetime.now()



from .crude_FR import CrudeFR
from .utils import delayed, delayed_with_discharge
from ..params import clinical
from ..utils import param_property, sliced


class HospitalizationWithDelay(CrudeFR):
    """
    Model in which infected can become hospitalized and suffer a constant
    hospitalization fatality rate.

    Attributes:
        severe_delay (float):
            Duration between symptom onset to hospitalization.
        critical_delay (float):
            Duration between symptom onset to ICU treatment.
    """

    params = clinical.DEFAULT

    # Primary parameters
    severe_delay: float = param_property(default=0.0)
    critical_delay: float = param_property(default=0.0)

    #
    # Data methods
    #
    def get_data_deaths(self, idx):
        K = max(self.K, 0)
        try:
            critical = self["critical_cases", idx]
        except KeyError:
            return delayed(self["severe_cases", idx] * self.HFR, self.severe_period, K)
        else:
            return delayed(critical * self.ICUFR, self.critical_period, K)

    def get_data_critical(self, idx):
        data = self["critical_cases"]
        period = self.critical_period
        data = delayed_with_discharge(data, 0, period, self.K, positive=True)
        return sliced(data, idx)

    def get_data_severe_cases(self, idx):
        K = max(self.K, 0)
        data = delayed(self["cases"] * self.Qsv, self.severe_delay, K)
        return sliced(data, idx)

    def get_data_critical_cases(self, idx):
        K = max(self.K, 0)
        values = self["severe_cases"] * (self.Qcr / self.Qsv)
        delay = self.severe_delay - self.critical_delay
        data = delayed(values, delay, K)
        return sliced(data, idx)



import numpy as np

from .base import formula

IGNORE = ("sigma", "rho", "prob_symptoms")
sir_formula = lambda ignore=(): formula("SIR", ignore=(*ignore, *IGNORE))


@sir_formula()
def R0(beta, gamma):
    """
    R0 from model parameters.
    """
    return beta / gamma


@sir_formula()
def beta(R0, gamma):
    """
    R0 from model parameters.
    """
    return R0 * gamma


@sir_formula()
def R0_from_K(K, gamma) -> float:
    """
    Return R0 from the exponential growth factor K and the other model
    parameters.
    """
    return 1.0 + K / gamma


@sir_formula()
def K(R0, gamma) -> float:
    """
    Return the exponential growth factor K from R0 and the other model
    parameters.
    """
    return gamma * (R0 - 1)


#
# Initial conditions
#
@sir_formula()
def time_to_seed(cases, R0, gamma) -> float:
    """
    Simple exponential extrapolation for the time from the first seed to
    reaching the given number of cases using the SIR model.

    This is a decent approximation for the outset of an epidemic. If the seed
    consists of more than a single individual, divide the number of cases by
    the estimated seed size.

    This extrapolation simply assumes that the I compartment grows with
    ``I = I0 * exp(K * t)`` and the number of cases as ``int(beta * I(t), t) + seed``
    """
    if R0 <= 1:
        raise ValueError(f"R0 must be greater than one (got {R0})")
    seed = 1
    return np.log((cases * (R0 - 1) + seed) / (2 * R0 - 1)) / (R0 - 1) / gamma


@sir_formula(ignore=["gamma"])
def infectious_from_cases(cases, R0) -> float:
    """
    Initializes the "infectious" component of a SIR model from the current
    number of cases.

    This formula assumes a perfect exponential growth.
    """
    if R0 <= 1:
        raise ValueError(f"R0 must be greater than one (got {R0})")

    seed = 1
    return (cases * (R0 - 1) + seed) / (2 * R0 - 1)


@sir_formula(ignore=["gamma"])
def state_from_cases(population, cases, R0):
    """
    Initialize the 3 components of a SIR model from cases and population.
    """
    I = infectious_from_cases.formula(cases, R0)  # noqa: E754
    R = cases - I
    S = population - I - R
    return np.array([S, I, R])



from .base import RegionProperty, cached
from .. import plot as plt
from ..utils import trim_weeks


class PyplotProperty(RegionProperty):
    __slots__ = ()

    @cached
    def weekday_rate(self, disease=None, **kwargs):
        """
        Show rate of notification by weekday.
        """
        data = self.region.pydemic.epidemic_curve(disease, diff=True)
        data = trim_weeks(data)
        return plt.weekday_rates(data, **kwargs)

    @cached
    def cases_and_deaths(self, disease=None, **kwargs):
        """
        Return a plot of cases and deaths
        """
        curves = self.region.pydemic.epidemic_curve(disease)
        kwargs.setdefault("tight_layout", True)
        return plt.cases_and_deaths(curves, **kwargs)



import locale
import os

import pytest


@pytest.yield_fixture
def en():
    en = "en_US.UTF-8"
    lc_all = locale.getlocale(locale.LC_ALL)
    lc_messages = locale.getlocale(locale.LC_MESSAGES)
    lang = os.environ.get("LANG")
    language = os.environ.get("LANGUAGE")

    try:
        locale.setlocale(locale.LC_ALL, en)
        locale.setlocale(locale.LC_MESSAGES, en)
        os.environ["LANG"] = en
        os.environ["LANGUAGE"] = en
        yield en
    except locale.Error:
        return pytest.skip("en_US.UTF-8 locale is not available in this machine")
    finally:
        locale.setlocale(locale.LC_ALL, lc_all)
        locale.setlocale(locale.LC_MESSAGES, lc_messages)

    if lang:
        os.environ["LANG"] = lang
    else:
        del os.environ["LANG"]

    if language:
        os.environ["LANGUAGE"] = language
    else:
        del os.environ["LANGUAGE"]



from collections.abc import Mapping
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import Model


class ParamsInfo(Mapping):
    """
    Object attached to WithParams subclasses that hold information about
    parameters.

    It behaves as a mapping from parameter names to their corresponding
    descriptors.
    """

    __slots__ = ("model", "primary", "alternative", "all")

    def __init__(self, cls: type):
        self.model = cls

        # Initialize parameter sets
        params = tuple(iter_params(self.model))
        self.primary = frozenset(k for k, v in params if not v.is_derived)
        self.alternative = frozenset(k for k, v in params if v.is_derived)
        self.all = self.primary | self.alternative

    def __contains__(self, key):
        return key in self.all

    def __getitem__(self, item):
        if item in self.all:
            return getattr(self.model, item)
        raise KeyError(item)

    def __len__(self) -> int:
        return len(self.all)

    def __iter__(self):
        return iter(self.all)

    def is_static(self, param: str, instance: "Model"):
        """
        Return True if parameter is static for model instance.
        """
        param = instance.get_param(param, param=True)
        return True


def iter_params(cls):
    """
    Iterate over all parameter descriptors in the given class.
    """
    for k in dir(cls):
        v = getattr(cls, k, None)
        if hasattr(v, "__get__") and getattr(v, "is_param", False):
            yield k, v



from collections import MutableSequence
from typing import TYPE_CHECKING
from weakref import ref

if TYPE_CHECKING:
    from .model_group import ModelGroup


class ModelList(MutableSequence):
    """
    Implements the "models" attribute of a model group.

    It offers a list-like interface to the models included in a ModelGroup.
    """

    __slots__ = ("_data", "_group")

    group: "ModelGroup" = property(lambda self: self._group())
    _data: list

    def __init__(self, group: "ModelGroup", data: list):
        self._group = ref(group)
        self._data = data

    def __getitem__(self, idx):
        if isinstance(idx, str):
            for m in self._data:
                if m.name == idx:
                    return m
            else:
                raise KeyError(f"no model named {idx!r}")

        data = self._data[idx]
        if isinstance(data, list):
            cls = type(self._group)
            return cls(data)
        return data

    def __setitem__(self, idx, obj) -> None:
        old = self[idx]
        model_cls = self.group.kind
        if isinstance(old, type(self._group)):
            data = getattr(obj, "models", obj)
            if not all(isinstance(m, model_cls) for m in data):
                raise self._insert_type_error()
            self._data[idx] = data
        elif isinstance(obj, model_cls):
            self._data[idx] = obj
        else:
            raise self._insert_type_error()

    def __delitem__(self, i: int) -> None:
        del self._data[i]

    def __len__(self) -> int:
        return len(self._data)

    def insert(self, index: int, object) -> None:
        if not isinstance(object, self.group.kind):
            raise self._insert_type_error()
        self._data[index] = object

    def _insert_type_error(self):
        return TypeError("can only insert model instances to model group")



class TestNamespaceModules:
    """
    Prevent accidentally deleting symbols from namespace modules.
    """

    def test_pydemic_all_has_the_correct_symbols(self):
        import pydemic.all as mod

        names = set(dir(mod))
        for name in ("mundi", "mdm", "mhc", "date"):
            assert name in names

    def test_pydemic_region_has_the_correct_symbols(self):
        import pydemic.region as mod

        names = set(dir(mod))
        for name in ("hospital_capacity", "icu_capacity", "population", "age_distribution"):
            assert name in names
